import datetime
import pathlib
import traceback
from abc import ABC, abstractmethod
from functools import cached_property
from logging import getLogger
from typing import Optional, Union

from competitive_verifier import git, github, log
from competitive_verifier.download.main import run_impl as run_download
from competitive_verifier.error import VerifierError
from competitive_verifier.models import (
    FileResult,
    ResultStatus,
    Verification,
    VerificationFile,
    VerificationInput,
    VerificationResult,
    VerifyCommandResult,
)
from competitive_verifier.resource import ulimit_stack
from competitive_verifier.verify.split_state import SplitState

logger = getLogger(__name__)


class InputContainer(ABC):
    input: VerificationInput
    verification_time: datetime.datetime
    prev_result: Optional[VerifyCommandResult]
    split_state: Optional[SplitState]

    def __init__(
        self,
        *,
        input: VerificationInput,
        verification_time: datetime.datetime,
        prev_result: Optional[VerifyCommandResult],
        split_state: Optional[SplitState],
    ) -> None:
        self.input = input
        self.verification_time = verification_time
        self.prev_result = prev_result
        self.split_state = split_state

    @abstractmethod
    def get_file_timestamp(self, path: pathlib.Path) -> datetime.datetime:
        ...

    def file_need_verification(
        self,
        path: pathlib.Path,
        file_result: FileResult,
    ) -> bool:
        if not path.exists():
            return False
        base_time = min(self.verification_time, self.get_file_timestamp(path))
        result = file_result.need_verification(base_time)
        if result:
            logger.info(
                "%s needs verification. base_time: %s", path.as_posix(), base_time
            )
        else:
            logger.info(
                "%s doesn't need verification. base_time: %s",
                path.as_posix(),
                base_time,
            )
        return result

    @cached_property
    def verification_files(self) -> dict[pathlib.Path, VerificationFile]:
        """
        List of verification files.
        """
        return {p: f for p, f in self.input.files.items() if f.is_verification()}

    @cached_property
    def skippable_verification_files(self) -> dict[pathlib.Path, VerificationFile]:
        return {
            p: f
            for p, f in self.verification_files.items()
            if f.is_skippable_verification()
        }

    @cached_property
    def remaining_verification_files(self) -> dict[pathlib.Path, VerificationFile]:
        """
        List of verification files that have not yet been verified.
        """
        verification_files = {
            p: f
            for p, f in self.verification_files.items()
            if p not in self.skippable_verification_files
        }

        if self.prev_result is None:
            return verification_files

        not_updated_files = set(
            k
            for k, v in self.input.filterd_files(self.prev_result.files)
            if not self.file_need_verification(k, v)
        )
        verification_files = {
            p: f for p, f in verification_files.items() if p not in not_updated_files
        }
        return verification_files

    @cached_property
    def current_verification_files(self) -> dict[pathlib.Path, VerificationFile]:
        """
        List of verification files that self should verify.

        if ``split_state`` is None the property is ``remaining_verification_files``;

        else ``split_state.split(remaining_verification_files)``.
        """
        if self.split_state is None:
            return self.remaining_verification_files

        lst = [(p, f) for p, f in self.remaining_verification_files.items()]
        lst.sort(key=lambda tup: tup[0])

        return {p: f for p, f in self.split_state.split(lst)}


class BaseVerifier(InputContainer):
    timeout: float
    default_tle: Optional[float]
    default_mle: Optional[float]
    split_state: Optional[SplitState]

    _result: Optional[VerifyCommandResult]

    def __init__(
        self,
        input: VerificationInput,
        *,
        timeout: float,
        default_tle: Optional[float],
        default_mle: Optional[float],
        prev_result: Optional[VerifyCommandResult],
        split_state: Optional[SplitState],
        verification_time: Optional[datetime.datetime] = None,
    ) -> None:
        super().__init__(
            input=input,
            verification_time=verification_time or self.now().astimezone(),
            prev_result=prev_result,
            split_state=split_state,
        )
        self._input = input
        self.timeout = timeout
        self.default_tle = default_tle
        self.default_mle = default_mle
        self._result = None

    @property
    def force_result(self) -> VerifyCommandResult:
        if self._result is None:
            raise VerifierError("Not verified yet.")
        return self._result

    @property
    def is_first(self) -> bool:
        if not self.split_state:
            return True
        return self.split_state.index == 0

    def now(self) -> datetime.datetime:
        return datetime.datetime.now(datetime.timezone.utc)

    def verify(self, *, download: bool = True) -> VerifyCommandResult:
        start_time = self.now()
        with log.group("current_verification_files"):
            current_verification_files = self.current_verification_files
        logger.info(
            "current_verification_files: %s",
            " ".join(p.as_posix() for p in current_verification_files.keys()),
        )
        try:
            ulimit_stack()
        except BaseException:
            logger.warning("failed to increase the stack size[ulimit]")

        if self.prev_result:
            file_results = {
                k: v.model_copy(update={"newest": False})
                for k, v in self.input.filterd_files(self.prev_result.files)
                if k.exists()
            }
        else:
            file_results = dict[pathlib.Path, FileResult]()

        for p, f in current_verification_files.items():
            logger.info("Start: %s", p.as_posix())

            def enumerate_verifications() -> list[VerificationResult]:
                logger.debug(repr(f))
                prev_time = self.now()
                verifications = list[VerificationResult]()
                try:
                    if download and not run_download(f, check=True, group_log=False):
                        raise Exception()
                except BaseException as e:
                    verifications.append(
                        self.create_command_result(ResultStatus.FAILURE, prev_time)
                    )
                    logger.exception("Failed to download", e)
                    return verifications

                for ve in f.verification_list:
                    logger.debug("command=%s", repr(ve))
                    prev_time = self.now()
                    if (prev_time - start_time).total_seconds() > self.timeout:
                        logger.warning("Skip[Timeout]: %s, %s", p, repr(ve))
                        verifications.append(
                            self.create_command_result(
                                ResultStatus.SKIPPED,
                                prev_time,
                                name=ve.name,
                            )
                        )
                        return verifications

                    try:
                        rs, error_message = self.run_verification(ve)
                        if error_message:
                            logger.error("%s: %s, %s", error_message, p, repr(ve))
                            if github.env.is_in_github_actions():
                                github.print_error(
                                    message=f"{error_message} {p.as_posix()}",
                                    file=str(p.resolve()),
                                )
                        verifications.append(
                            self.create_command_result(rs, prev_time, name=ve.name)
                        )
                    except BaseException as e:
                        message = (
                            e.message
                            if isinstance(e, VerifierError)
                            else "Failed to verify"
                        )
                        logger.error("%s: %s, %s", message, p, repr(ve))
                        traceback.print_exc()
                        verifications.append(
                            self.create_command_result(
                                ResultStatus.FAILURE,
                                prev_time,
                                name=ve.name,
                            )
                        )
                return verifications

            with log.group(f"Verify: {p.as_posix()}"):
                file_results[p] = FileResult(verifications=enumerate_verifications())

        sippable_file_results = self.skippable_results()
        self._result = VerifyCommandResult(
            total_seconds=(self.now() - start_time).total_seconds(),
            files=file_results | sippable_file_results,
        )
        return self._result

    def run_verification(
        self, verification: Verification
    ) -> tuple[Union[ResultStatus, VerificationResult], Optional[str]]:
        """Run verification

        Returns:
            tuple[ResultStatus, Optional[str]]: (Result, error_message)
        """
        error_message: Optional[str] = None
        if not verification.run_compile_command():
            error_message = "Failed to compile"

        rs = verification.run(self)

        if rs.status != ResultStatus.SUCCESS:
            error_message = "Failed to test"
        return (rs, error_message)

    def skippable_results(self) -> dict[pathlib.Path, FileResult]:
        """
        Run skippable verification
        """
        results = dict[pathlib.Path, FileResult]()
        if self.is_first:
            for p, f in self.skippable_verification_files.items():
                logger.info("Start skippable: %s", p.as_posix())
                verifications = list[VerificationResult]()
                prev_time = self.now()

                for v in f.verification_list:
                    rs = self.run_verification(v)[0]
                    verifications.append(
                        self.create_command_result(rs, prev_time, name=v.name)
                    )
                results[p] = FileResult(
                    verifications=verifications,
                    newest=True,
                )
        return results

    def create_command_result(
        self,
        status_or_result: Union[ResultStatus, VerificationResult],
        prev_time: datetime.datetime,
        *,
        name: Optional[str] = None,
    ) -> VerificationResult:
        if isinstance(status_or_result, VerificationResult):
            return status_or_result

        elapsed = (self.now() - prev_time).total_seconds()
        return VerificationResult(
            verification_name=name,
            status=status_or_result,
            elapsed=elapsed,
            last_execution_time=self.verification_time,
        )


class Verifier(BaseVerifier):
    use_git_timestamp: bool

    def __init__(
        self,
        input: VerificationInput,
        *,
        timeout: float,
        default_tle: Optional[float],
        default_mle: Optional[float],
        prev_result: Optional[VerifyCommandResult],
        split_state: Optional[SplitState],
        verification_time: Optional[datetime.datetime] = None,
        use_git_timestamp: bool,
    ) -> None:
        super().__init__(
            input=input,
            verification_time=verification_time or self.now().astimezone(),
            prev_result=prev_result,
            split_state=split_state,
            timeout=timeout,
            default_tle=default_tle,
            default_mle=default_mle,
        )
        self.use_git_timestamp = use_git_timestamp

    def get_file_timestamp(self, path: pathlib.Path) -> datetime.datetime:
        if self.use_git_timestamp:
            return git.get_commit_time(self.input.transitive_depends_on[path])
        else:
            dependicies = self.input.transitive_depends_on[path]

            timestamp = max(x.stat().st_mtime for x in dependicies)
            system_local_timezone = self.now().astimezone().tzinfo
            return datetime.datetime.fromtimestamp(
                timestamp, tz=system_local_timezone
            ).replace(
                microsecond=0
            )  # microsecond=0 is required because it's erased in git commit



import argparse
import logging
import sys
from itertools import chain
from logging import getLogger
from typing import Iterable, Optional, Union

from competitive_verifier import oj
from competitive_verifier.arg import (
    add_verbose_argument,
    add_verify_files_json_argument,
)
from competitive_verifier.error import VerifierError
from competitive_verifier.log import configure_stderr_logging
from competitive_verifier.models import (
    ProblemVerification,
    VerificationFile,
    VerificationInput,
)
from competitive_verifier.resource import ulimit_stack

logger = getLogger(__name__)

UrlOrVerificationFile = Union[str, VerificationFile]


def parse_urls(
    input: Union[UrlOrVerificationFile, Iterable[UrlOrVerificationFile]]
) -> set[str]:
    def parse_single(url_or_file: UrlOrVerificationFile) -> Iterable[str]:
        if isinstance(url_or_file, str):
            return (url_or_file,)
        else:
            return enumerate_urls(url_or_file)

    if isinstance(input, (str, VerificationFile)):
        return set(parse_single(input))

    return set(chain.from_iterable(map(parse_single, input)))


def enumerate_urls(file: VerificationFile) -> Iterable[str]:
    for v in file.verification_list:
        if isinstance(v, ProblemVerification):
            yield v.problem


def run_impl(
    input: Union[UrlOrVerificationFile, Iterable[UrlOrVerificationFile]],
    check: bool = False,
    group_log: bool = False,
) -> bool:
    result = True
    try:
        ulimit_stack()
    except Exception:
        logger.warning("failed to increase the stack size[ulimit]")
    for url in parse_urls(input):
        if not oj.download(url, group_log=group_log):
            result = False

    if check and not result:
        raise VerifierError("Failed to download")
    return result


def run(args: argparse.Namespace) -> bool:
    default_level = logging.INFO
    if args.verbose:
        default_level = logging.DEBUG
    configure_stderr_logging(default_level)

    logger.debug("arguments=%s", vars(args))
    logger.info("verify_files_json=%s", str(args.verify_files_json))
    logger.info("urls=%s", args.urls)
    files: list[VerificationFile] = []
    if args.verify_files_json:
        verification = VerificationInput.parse_file_relative(args.verify_files_json)
        files = list(verification.files.values())

    return run_impl(files + args.urls, group_log=True)


def argument(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    add_verbose_argument(parser)
    add_verify_files_json_argument(parser, required=False)
    parser.add_argument(
        "urls",
        nargs="*",
        help="A list of problem URL",
    )
    return parser


def main(args: Optional[list[str]] = None) -> None:
    try:
        parsed = argument(argparse.ArgumentParser()).parse_args(args)
        if not run(parsed):
            sys.exit(1)
    except Exception as e:
        sys.stderr.write(str(e))
        sys.exit(2)


if __name__ == "__main__":
    main()



import pathlib
from typing import Annotated, BinaryIO, Literal, Optional, Union

import yaml
from pydantic import BaseModel, ConfigDict, Field

from competitive_verifier.models import DocumentOutputMode, ForcePosixPath

from .render_data import IndexRenderData, MultiCodePageData, PageRenderData

_separator: bytes = b"---"


class FrontMatter(BaseModel):
    model_config = ConfigDict(extra="allow")

    display: Optional[DocumentOutputMode] = None
    title: Optional[str] = None
    layout: Union[
        Literal["toppage"],
        Literal["document"],
        Literal["multidoc"],
        str,
        None,
    ] = None
    documentation_of: Union[str, Annotated[list[str], Field(min_length=1)], None] = None
    keep_single: Optional[bool] = None
    data: Union[PageRenderData, MultiCodePageData, IndexRenderData, None] = None
    redirect_from: Optional[list[str]] = None
    """For jekyll-redirect-from plugin
    """
    redirect_to: Optional[str] = None
    """For jekyll-redirect-from plugin
    """

    def model_dump_yml(self):
        d = self.model_dump(
            mode="json",
            by_alias=True,
            exclude_none=True,
        )
        return yaml.safe_dump(d, encoding="utf-8", line_break="\n")


class Markdown(BaseModel):
    path: Optional[ForcePosixPath] = None
    front_matter: Optional[FrontMatter]
    content: bytes

    @classmethod
    def make_default(cls, source_path: pathlib.Path):
        return cls(
            front_matter=FrontMatter(documentation_of=source_path.as_posix()),
            content=b"",
        )

    @classmethod
    def load_file(cls, path: pathlib.Path):
        with path.open("rb") as fp:
            return cls.load(fp, path)

    @classmethod
    def load(cls, fp: BinaryIO, path: Optional[pathlib.Path] = None):
        front_matter, content = split_front_matter(fp.read())
        return Markdown(
            path=path,
            front_matter=front_matter,
            content=content,
        )

    def dump_merged(self, fp: BinaryIO):
        merge_front_matter(
            fp,
            front_matter=self.front_matter,
            content=self.content,
        )


def split_front_matter_raw(content: bytes) -> tuple[Optional[bytes], bytes]:
    lines = content.splitlines()
    if len(lines) == 0 or lines[0].rstrip() != _separator:
        return (None, content)
    for i, line in enumerate(lines):
        if i == 0:
            continue
        if line.rstrip() == _separator:
            break
    else:
        return None, content

    front_matter = b"\n".join(lines[1:i])
    content = b"\n".join(lines[i + 1 :])
    return front_matter, content


def split_front_matter(content: bytes) -> tuple[Optional[FrontMatter], bytes]:
    fm_bytes, content = split_front_matter_raw(content)
    if fm_bytes is None:
        return None, content
    fy = yaml.safe_load(fm_bytes)
    if fy:
        return FrontMatter.model_validate(fy), content
    else:
        return FrontMatter(), content


def merge_front_matter(
    fp: BinaryIO,
    *,
    front_matter: Optional[FrontMatter],
    content: bytes,
):
    if front_matter:
        fp.write(_separator)
        fp.write(b"\n")
        fp.write(front_matter.model_dump_yml())
        fp.write(_separator)
        fp.write(b"\n")
    fp.write(content)



# Python Version: 3.x
import functools
import pathlib
import re
from logging import getLogger
from typing import Iterable, Mapping

from competitive_verifier.oj.verify.utils import read_text_normalized

logger = getLogger(__name__)


# special comments like Vim and Python: see https://www.python.org/dev/peps/pep-0263/
@functools.lru_cache(maxsize=None)
def list_special_comments(path: pathlib.Path) -> Mapping[str, str]:
    pattern = re.compile(
        r"\b(?:verify-helper|verification-helper|competitive-verifier):\s*([0-9A-Za-z_]+)(?:\s(.*))?$"
    )
    attributes: dict[str, str] = {}
    for line in read_text_normalized(path).splitlines():
        matched = pattern.search(line)
        if matched:
            key = matched.group(1)
            value = (matched.group(2) or "").strip()
            attributes[key] = value
    return attributes


@functools.lru_cache(maxsize=None)
def list_embedded_urls(path: pathlib.Path) -> Iterable[str]:
    pattern = re.compile(
        r"""['"`]?https?://\S*"""
    )  # use a broad pattern. There are no needs to make match strict.
    content = read_text_normalized(path)
    urls: list[str] = []
    for url in pattern.findall(content):
        # The URL may be written like `"https://atcoder.jp/"`. In this case, we need to remove `"`s around the URL.
        # We also need to remove trailing superfluous chars in a case like `{"url":"https://atcoder.jp/"}`.
        for quote in ("'", '"', "`"):
            if url.startswith(quote):
                end_quote_pos = url.rfind(quote)
                if end_quote_pos == 0:
                    # Remove opening quote from the URL like `"https://atcoder.jp/`
                    url = url[1:]
                else:
                    # Remove quotes and trailing superfluous chars around the URL
                    url = url[1:end_quote_pos]
                break
        urls.append(url)
    return sorted(set(urls))



# The file is inspired by Tyrrrz/GitHubActionsTestLogger
# https://github.com/Tyrrrz/GitHubActionsTestLogger/blob/04fe7796a047dbd0e3cd6a46339b2a50f5125317/GitHubActionsTestLogger/TestSummary.cs

import os
import pathlib
from collections import Counter
from itertools import chain
from typing import Optional, TextIO

from competitive_verifier.models import (
    FileResult,
    JudgeStatus,
    ResultStatus,
    TestcaseResult,
    VerifyCommandResult,
)

SUCCESS = ResultStatus.SUCCESS
FAILURE = ResultStatus.FAILURE
SKIPPED = ResultStatus.SKIPPED


def to_human_str_seconds(total_seconds: float) -> str:
    hours = int(total_seconds // 3600)
    rm = total_seconds % 3600
    minutes = int(rm // 60)
    rm %= 60
    seconds = rm

    if hours > 0:
        return f"{hours}h {minutes}m"
    elif minutes > 0:
        return f"{minutes}m {int(seconds)}s"
    elif total_seconds >= 10:
        return f"{int(seconds)}s"
    elif total_seconds > 1:
        return f"{total_seconds:.1f}s"
    else:
        return f"{int(total_seconds*1000)}ms"


def to_human_str_mega_bytes(total_mega_bytes: float) -> str:
    if total_mega_bytes < 0.001:
        return "0MB"
    if total_mega_bytes < 100:
        return f"{total_mega_bytes:.3g}MB"
    return f"{int(total_mega_bytes)}MB"


class TableWriter:
    def __init__(self, fp: TextIO, header: list[str]) -> None:
        self.size = len(header)
        self.fp = fp
        self.write_table_line(*header)

    def write_table_line(self, *cells: str) -> None:
        assert len(cells) == self.size
        fp = self.fp
        for c in cells:
            fp.write("|")
            fp.write(c)
        fp.write("|\n")


def write_summary(fp: TextIO, result: VerifyCommandResult):
    def with_icon(icon: str, text: str) -> str:
        return icon + "&nbsp;&nbsp;" + text

    file_results: list[tuple[pathlib.Path, FileResult]] = []
    past_results: list[tuple[pathlib.Path, FileResult]] = []
    for p, fr in result.files.items():
        if fr.newest:
            file_results.append((p, fr))
        else:
            past_results.append((p, fr))

    file_results.sort(key=lambda t: t[0])
    past_results.sort(key=lambda t: t[0])
    counter = Counter(
        r.status for r in chain.from_iterable(f[1].verifications for f in file_results)
    )

    fp.write("# ")

    if counter.get(FAILURE):
        emoji_status = "❌"
    elif counter.get(SKIPPED):
        emoji_status = "⚠"
    else:
        emoji_status = "✔"

    fp.write(emoji_status)
    fp.write(" ")
    fp.write(os.getenv("COMPETITIVE_VERIFY_SUMMARY_TITLE", "Verification result"))
    fp.write("\n\n")

    fp.write("- ")
    fp.write(with_icon("✔", "All test case results are `success`"))
    fp.write("\n")
    fp.write("- ")
    fp.write(with_icon("❌", "Test case results containts `failure`"))
    fp.write("\n")
    fp.write("- ")
    fp.write(with_icon("⚠", "Test case results containts `skipped`"))
    fp.write("\n\n\n")

    header = [
        with_icon("📝", "File"),
        "✔<br>Passed",
        "❌<br>Failed",
        "⚠<br>Skipped",
        "∑<br>Total",
        "⏳<br>Elapsed",
        "🦥<br>Slowest",
        "🐘<br>Heaviest",
    ]
    alignment = [":---"] + [":---:"] * (len(header) - 1)

    def write_table_file_result(results: list[tuple[pathlib.Path, FileResult]]) -> None:
        for p, fr in results:
            counter = Counter(r.status for r in fr.verifications)
            if counter.get(FAILURE):
                emoji_status = "❌"
            elif counter.get(SKIPPED):
                emoji_status = "⚠"
            else:
                emoji_status = "✔"
            elapsed = sum(r.elapsed for r in fr.verifications)
            slowest = max(
                (r.slowest for r in fr.verifications if r.slowest is not None),
                default=None,
            )
            heaviest = max(
                (r.heaviest for r in fr.verifications if r.heaviest is not None),
                default=None,
            )
            tb.write_table_line(
                with_icon(emoji_status, p.as_posix()),
                str(counter.get(SUCCESS, "-")),
                str(counter.get(FAILURE, "-")),
                str(counter.get(SKIPPED, "-")),
                str(sum(counter.values())),
                to_human_str_seconds(elapsed),
                "-" if slowest is None else to_human_str_seconds(slowest),
                "-" if heaviest is None else to_human_str_mega_bytes(heaviest),
            )

    if file_results:
        fp.write("## Results\n")
        tb = TableWriter(fp, header)
        tb.write_table_line(*alignment)
        tb.write_table_line(
            "_**Sum**_",
            str(counter.get(SUCCESS, "-")),
            str(counter.get(FAILURE, "-")),
            str(counter.get(SKIPPED, "-")),
            str(sum(counter.values())),
            to_human_str_seconds(result.total_seconds),
            "-",
            "-",
        )
        tb.write_table_line(*[""] * len(header))

        write_table_file_result(file_results)

    if past_results:
        fp.write("## Past results\n")
        tb = TableWriter(fp, header)
        tb.write_table_line(*alignment)
        write_table_file_result(past_results)

    if counter.get(FAILURE):
        assert file_results
        first_failure = True
        for p, fr in file_results:
            cases = [
                DisplayTestcaseResult(
                    environment=v.verification_name, **(c.model_dump())
                )
                for v in fr.verifications
                for c in (v.testcases or [])
                if c.status != JudgeStatus.AC
            ]
            if not cases:
                continue
            if first_failure:
                fp.write("## Failed tests\n\n")
                first_failure = False
            fp.write(f"### {p.as_posix()}\n\n")

            etb = TableWriter(fp, ["env", "name", "status", "Elapsed", "Memory"])
            etb.write_table_line(*[":---"] * 2 + [":---:"] * 3)

            for c in cases:
                etb.write_table_line(
                    c.environment or "",
                    c.name,
                    c.status.name,
                    c.elapsed_str,
                    c.memory_str,
                )


class DisplayTestcaseResult(TestcaseResult):
    environment: Optional[str]

    @property
    def elapsed_str(self):
        return to_human_str_seconds(self.elapsed)

    @property
    def memory_str(self):
        return to_human_str_mega_bytes(self.memory) if self.memory else "-"



import pathlib
from typing import Protocol

from pydantic import BaseModel


class FilePaths(BaseModel):
    root: pathlib.Path
    dest_root: pathlib.Path


class ConfigDirSetter(Protocol):
    def __call__(
        self,
        name: str,
    ) -> pathlib.Path:
        ...



import contextlib
import json
import os
import pathlib
import platform
import shlex
import signal
import sys
import tempfile
import threading
import time
from logging import getLogger
from subprocess import PIPE, Popen, TimeoutExpired
from typing import Annotated, Any, BinaryIO, Optional, Union

import onlinejudge_command.format_utils as fmtutils
import onlinejudge_command.pretty_printers as pretty_printers
import onlinejudge_command.utils as utils
from onlinejudge_command.subcommand.test import (
    CompareMode,
    DisplayMode,
    JudgeStatus,
    build_match_function,
)
from onlinejudge_command.subcommand.test import check_gnu_time as orig_check_gnu_time
from onlinejudge_command.subcommand.test import run_checking_output
from pydantic import BaseModel, Field
from pydantic.functional_validators import BeforeValidator

from competitive_verifier.models import ResultStatus, TestcaseResult, VerificationResult

from .func import checker_exe_name, get_cache_directory, get_directory

logger = getLogger(__name__)


class OjTestArguments(BaseModel):
    """onlinejudge_command.subcommand.test.add_subparser"""

    command: Union[str, list[str]]
    cookie: pathlib.Path
    env: Optional[dict[str, str]] = None
    directory: pathlib.Path
    judge: Optional[pathlib.Path]
    tle: Optional[float]
    mle: Optional[float]
    error: Optional[float]
    print_input: bool = True
    format: str = "%s.%e"
    test: list[pathlib.Path] = Field(default_factory=list)
    gnu_time: Optional[str] = None
    log_file: Optional[pathlib.Path] = None
    silent: bool = False
    ignore_backup: bool = True
    display_mode: DisplayMode = DisplayMode.SUMMARY
    compare_mode: CompareMode = CompareMode.CRLF_INSENSITIVE_EXACT_MATCH


class OjExecInfo(BaseModel):
    answer: Optional[bytes]
    elapsed: float
    memory: Optional[float]


def oj_exec_command(
    command: Union[str, list[str]],
    *,
    env: Optional[dict[str, str]],
    stdin: Optional[BinaryIO] = None,
    input: Optional[bytes] = None,
    timeout: Optional[float] = None,
    gnu_time: Optional[str] = None,
) -> tuple[OjExecInfo, Popen[bytes]]:
    if input is not None:
        assert stdin is None
        stdin = PIPE  # type: ignore
    if gnu_time is not None:
        context: Any = tempfile.NamedTemporaryFile(delete=True)
    else:
        context = contextlib.nullcontext()
    with context as fh:
        if isinstance(command, str):
            command_str = command
            command = shlex.split(command)
            if gnu_time is not None:
                command = [gnu_time, "-f", "%M", "-o", fh.name, "--"] + command
            if sys.platform == "win32":
                # HACK: without this encoding and decoding, something randomly fails with multithreading; see https://github.com/kmyk/online-judge-tools/issues/468
                command = command_str.encode().decode()  # type: ignore
        begin = time.perf_counter()

        # We need kill processes called from the "time" command using process groups. Without this, orphans spawn. see https://github.com/kmyk/online-judge-tools/issues/640
        preexec_fn = None
        if gnu_time is not None and os.name == "posix":
            preexec_fn = os.setsid

        try:
            if env:
                env = os.environ | env
            proc = Popen(
                command,
                env=env,
                stdin=stdin,
                stdout=PIPE,
                stderr=sys.stderr,
                preexec_fn=preexec_fn,
            )  # pylint: disable=subprocess-popen-preexec-fn
        except FileNotFoundError:
            logger.error("No such file or directory: %s", command)
            sys.exit(1)
        except PermissionError:
            logger.error("Permission denied: %s", command)
            sys.exit(1)
        answer: Optional[bytes] = None
        try:
            answer, _ = proc.communicate(input=input, timeout=timeout)
        except TimeoutExpired:
            pass
        finally:
            if preexec_fn is not None:
                try:
                    if sys.platform != "win32":
                        os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
                except ProcessLookupError:
                    pass
            else:
                proc.terminate()

        end = time.perf_counter()
        memory: Optional[float] = None
        if gnu_time is not None:
            with open(fh.name) as fh1:
                reported = fh1.read()
            logger.debug("GNU time says:\n%s", reported)
            if reported.strip() and reported.splitlines()[-1].isdigit():
                memory = int(reported.splitlines()[-1]) / 1000
    return (
        OjExecInfo(
            answer=answer,
            memory=memory,
            elapsed=end - begin,
        ),
        proc,
    )


# flake8: noqa: C901
def display_result(
    proc: Popen[bytes],
    answer: str,
    memory: Optional[float],
    test_input_path: pathlib.Path,
    test_output_path: Optional[pathlib.Path],
    *,
    mle: Optional[float],
    display_mode: DisplayMode,
    compare_mode: CompareMode,
    does_print_input: bool,
    silent: bool,
    match_result: Optional[bool],
) -> JudgeStatus:
    """display_result prints the result of the test and its statistics.

    This function prints many logs and does some I/O.
    """

    # prepare the function to print the input
    is_input_printed = False

    def print_input() -> None:
        nonlocal is_input_printed
        if does_print_input and not is_input_printed:
            is_input_printed = True
            with test_input_path.open("rb") as inf:
                logger.info(
                    utils.NO_HEADER + "input:\n%s",
                    pretty_printers.make_pretty_large_file_content(
                        inf.read(), limit=40, head=20, tail=10
                    ),
                )

    # check TLE, RE or not
    status = JudgeStatus.AC
    if proc.returncode is None:
        logger.info(utils.FAILURE + "" + utils.red("TLE"))
        status = JudgeStatus.TLE
        if not silent:
            print_input()
    elif memory is not None and mle is not None and memory > mle:
        logger.info(utils.FAILURE + "" + utils.red("MLE"))
        status = JudgeStatus.MLE
        if not silent:
            print_input()
    elif proc.returncode != 0:
        logger.info(
            utils.FAILURE + "" + utils.red("RE") + ": return code %d", proc.returncode
        )
        status = JudgeStatus.RE
        if not silent:
            print_input()

    # check WA or not
    # 元の実装では TLE や RE でもこっちに来てしまうので elif に変更
    elif match_result is not None and not match_result:
        if status == JudgeStatus.AC:
            logger.info(utils.FAILURE + "" + utils.red("WA"))
        status = JudgeStatus.WA
        if not silent:
            print_input()
            if test_output_path is not None:
                with test_output_path.open("rb") as outf:
                    expected = outf.read().decode()
            else:
                expected = ""
            if display_mode == DisplayMode.SUMMARY:
                logger.info(
                    utils.NO_HEADER + "output:\n%s",
                    pretty_printers.make_pretty_large_file_content(
                        answer.encode(), limit=40, head=20, tail=10
                    ),
                )
                logger.info(
                    utils.NO_HEADER + "expected:\n%s",
                    pretty_printers.make_pretty_large_file_content(
                        expected.encode(), limit=40, head=20, tail=10
                    ),
                )
            elif display_mode == DisplayMode.ALL:
                logger.info(
                    utils.NO_HEADER + "output:\n%s",
                    pretty_printers.make_pretty_all(answer.encode()),
                )
                logger.info(
                    utils.NO_HEADER + "expected:\n%s",
                    pretty_printers.make_pretty_all(expected.encode()),
                )
            elif display_mode == DisplayMode.DIFF:
                logger.info(
                    utils.NO_HEADER
                    + pretty_printers.make_pretty_diff(
                        answer.encode(),
                        expected=expected,
                        compare_mode=compare_mode,
                        limit=40,
                    )
                )
            elif display_mode == DisplayMode.DIFF_ALL:
                logger.info(
                    utils.NO_HEADER
                    + pretty_printers.make_pretty_diff(
                        answer.encode(),
                        expected=expected,
                        compare_mode=compare_mode,
                        limit=-1,
                    )
                )
            else:
                assert False
    if match_result is None:
        if not silent:
            print_input()
            logger.info(
                utils.NO_HEADER + "output:\n%s",
                pretty_printers.make_pretty_large_file_content(
                    answer.encode(), limit=40, head=20, tail=10
                ),
            )
    if status == JudgeStatus.AC:
        logger.info(utils.SUCCESS + "" + utils.green("AC"))

    return status


class TestCase(BaseModel):
    name: str
    input: pathlib.Path
    output: Optional[pathlib.Path] = None


class OjTestcaseResult(BaseModel):
    status: JudgeStatus
    elapsed: float
    memory: Optional[float] = None
    exitcode: Annotated[
        Optional[int], BeforeValidator(lambda v: v if isinstance(v, int) else None)
    ]
    testcase: TestCase


class OjTestResult(BaseModel):
    is_success: bool

    elapsed: float

    slowest: float
    """max time [seconds]
    """

    heaviest: float
    """max memory [MB]
    """

    testcases: list[OjTestcaseResult]


def get_gnu_time_command() -> str:
    if platform.system() == "Darwin":
        return "gtime"
    else:
        return "time"


def check_gnu_time(gnu_time: Optional[str] = None) -> bool:
    return orig_check_gnu_time(gnu_time or get_gnu_time_command())


def test_single_case(
    test_name: str,
    test_input_path: pathlib.Path,
    test_output_path: Optional[pathlib.Path],
    *,
    lock: Optional[threading.Lock] = None,
    args: OjTestArguments,
) -> dict[str, Any]:
    # print the header earlier if not in parallel
    if lock is None:
        logger.info("")
        logger.info("%s", test_name)

    # run the binary
    with test_input_path.open("rb") as inf:
        info, proc = oj_exec_command(
            args.command,
            env=args.env,
            stdin=inf,
            timeout=args.tle,
            gnu_time=args.gnu_time,
        )
        # TODO: the `answer` should be bytes, not str
        answer: str = (info.answer or b"").decode(errors="replace")
        elapsed: float = info.elapsed
        memory: Optional[float] = info.memory

    # lock is require to avoid mixing logs if in parallel
    with lock or contextlib.nullcontext():
        if lock is not None:
            logger.info("")
            logger.info("%s", test_name)
        logger.info("time: %f sec", elapsed)
        if memory:
            logger.info("memory: %f MB", memory)

        match_function = build_match_function(
            compare_mode=CompareMode(args.compare_mode),
            error=args.error,
            judge_command=str(args.judge) if args.judge else None,
            silent=args.silent,
            test_input_path=test_input_path,
            test_output_path=test_output_path,
        )
        match_result = run_checking_output(
            answer=answer.encode(),
            test_output_path=test_output_path,
            is_special_judge=args.judge is not None,
            match_function=match_function,
        )
        status = display_result(
            proc,
            answer,
            memory,
            test_input_path,
            test_output_path,
            mle=args.mle,
            display_mode=DisplayMode(args.display_mode),
            compare_mode=CompareMode(args.compare_mode),
            does_print_input=args.print_input,
            silent=args.silent,
            match_result=match_result,
        )

    # return the result
    testcase = {
        "name": test_name,
        "input": str(test_input_path.resolve()),
    }
    if test_output_path:
        testcase["output"] = str(test_output_path.resolve())
    return {
        "status": status.value,
        "testcase": testcase,
        "output": answer,
        "exitcode": proc.returncode,
        "elapsed": elapsed,
        "memory": memory,
    }


def run(args: OjTestArguments) -> OjTestResult:
    # list tests
    if not args.test:
        args.test = fmtutils.glob_with_format(args.directory, args.format)  # by default
    if args.ignore_backup:
        args.test = fmtutils.drop_backup_or_hidden_files(args.test)
    tests = fmtutils.construct_relationship_of_files(
        args.test, args.directory, args.format
    )

    # check wheather GNU time is available
    if args.gnu_time is None:
        if platform.system() == "Darwin":
            args.gnu_time = "gtime"
        else:
            args.gnu_time = "time"
    if not check_gnu_time(args.gnu_time):
        logger.warning("GNU time is not available: %s", args.gnu_time)
        if platform.system() == "Darwin":
            logger.info(
                utils.HINT + "You can install GNU time with: $ brew install gnu-time"
            )
        args.gnu_time = None
    if args.mle is not None and args.gnu_time is None:
        raise RuntimeError("--mle is used but GNU time does not exist")

    # run tests
    history: list[OjTestcaseResult] = []
    for name, paths in sorted(tests.items()):
        history.append(
            OjTestcaseResult.model_validate(
                test_single_case(name, paths["in"], paths.get("out"), args=args),
            )
        )

    # summarize
    elapsed: float = 0.0
    slowest: float = -1.0
    slowest_name = ""
    heaviest: float = -1.0
    heaviest_name = ""
    ac_count = 0
    for result in history:
        elapsed += result.elapsed
        if result.status == JudgeStatus.AC:
            ac_count += 1
        if slowest < result.elapsed:
            slowest = result.elapsed
            slowest_name = result.testcase.name
        if result.memory is not None and heaviest < result.memory:
            heaviest = result.memory
            heaviest_name = result.testcase.name

    # print the summary
    logger.info("")
    logger.info("slowest: %f sec  (for %s)", slowest, slowest_name)
    if heaviest >= 0:
        logger.info("max memory: %f MB  (for %s)", heaviest, heaviest_name)
    if ac_count == len(tests):
        logger.info(
            utils.SUCCESS + "test " + utils.green("success") + ": %d cases", len(tests)
        )
    else:
        logger.info(
            utils.FAILURE + "test " + utils.red("failed") + ": %d AC / %d cases",
            ac_count,
            len(tests),
        )

    if args.log_file:
        with args.log_file.open(mode="w") as fh:
            json.dump(history, fh)

    # return the result
    return OjTestResult(
        is_success=ac_count == len(tests),
        slowest=slowest,
        elapsed=elapsed,
        heaviest=heaviest,
        testcases=history,
    )


def run_wrapper(
    *,
    url: str,
    command: Union[str, list[str]],
    env: Optional[dict[str, str]],
    tle: Optional[float],
    mle: Optional[float],
    error: Optional[float],
) -> VerificationResult:
    directory = get_directory(url)
    test_directory = directory / "test"

    checker_path: Optional[pathlib.Path] = directory / checker_exe_name
    if not checker_path.exists():
        checker_path = None

    args = OjTestArguments(
        command=command,
        env=env,
        cookie=get_cache_directory() / "cookie.txt",
        directory=test_directory,
        tle=tle,
        mle=mle,
        error=error,
        print_input=True,
        judge=checker_path,
    )
    result = run(args)

    return VerificationResult(
        status=ResultStatus.SUCCESS if result.is_success else ResultStatus.FAILURE,
        elapsed=result.elapsed,
        slowest=result.slowest,
        heaviest=result.heaviest,
        testcases=[
            TestcaseResult(
                name=case.testcase.name,
                elapsed=case.elapsed,
                memory=case.memory,
                status=case.status,
            )
            for case in result.testcases
        ],
    )



from typing import Optional

from pydantic import Field

from competitive_verifier.models import ShellCommand, ShellCommandLike
from competitive_verifier.oj.verify.languages.user_defined import UserDefinedLanguage
from competitive_verifier.oj.verify.models import OjVerifyUserDefinedConfig


class OjVerifyGoConfig(OjVerifyUserDefinedConfig):
    execute: ShellCommandLike = Field(
        default_factory=lambda: ShellCommand(
            command=["go", "run", "{basedir}/{path}"],
            env={"GO111MODULE": "off"},
        ),
    )


class GoLanguage(UserDefinedLanguage):
    def __init__(self, *, config: Optional[OjVerifyGoConfig]):
        super().__init__(extension="go", config=config or OjVerifyGoConfig())



import datetime
import enum
import pathlib
from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from logging import getLogger
from typing import AbstractSet, BinaryIO, Iterable, Optional

from pydantic import BaseModel

import competitive_verifier.git as git
import competitive_verifier.log as log
from competitive_verifier.models import (
    DocumentOutputMode,
    ForcePosixPath,
    ProblemVerification,
    ResultStatus,
    SortedPathSet,
    VerificationFile,
    VerificationInput,
    VerificationResult,
    VerifyCommandResult,
)
from competitive_verifier.util import normalize_bytes_text, read_text_normalized

from .config import ConfigYaml
from .front_matter import FrontMatter, Markdown
from .render_data import (
    CategorizedIndex,
    CodePageData,
    Dependency,
    EmbeddedCode,
    EnvTestcaseResult,
    IndexFiles,
    IndexRenderData,
    MultiCodePageData,
    PageRenderData,
    RenderLink,
    StatusIcon,
)

logger = getLogger(__name__)


def resolve_documentation_of(
    documentation_of: str,
    *,
    basedir: pathlib.Path,
) -> Optional[pathlib.Path]:
    def inner():
        if documentation_of.startswith("."):
            # a relative path
            path = basedir / pathlib.Path(documentation_of)
            if path.exists():
                return path
        elif documentation_of.startswith("//"):
            # from the document root
            path = pathlib.Path(documentation_of[2:])
            if path.exists():
                return path

        path = pathlib.Path(documentation_of)
        if path.exists():
            return path

        path = basedir / pathlib.Path(documentation_of)
        if path.exists():
            return path

    path = inner()
    if path:
        path = path.resolve()
        if path.is_relative_to(pathlib.Path.cwd()):
            return path.relative_to(pathlib.Path.cwd())
    return None


def _paths_to_render_links(
    paths: SortedPathSet, page_jobs: dict[pathlib.Path, "PageRenderJob"]
) -> list[RenderLink]:
    def get_link(path: pathlib.Path) -> Optional[RenderLink]:
        job = page_jobs.get(path)
        if not job:
            return None
        return job.to_render_link()

    return [
        link
        for link in map(
            get_link, sorted(paths, key=lambda p: str.casefold(p.as_posix()))
        )
        if link
    ]


class MultiTargetMarkdown(Markdown):
    path: ForcePosixPath  # pyright: ignore
    front_matter: FrontMatter  # pyright: ignore
    multi_documentation_of: list[pathlib.Path]


@dataclass
class UserMarkdowns:
    single: dict[pathlib.Path, Markdown]
    multi: list[MultiTargetMarkdown]

    @staticmethod
    def select_markdown(sources: set[pathlib.Path]) -> "UserMarkdowns":
        single: dict[pathlib.Path, Markdown] = {}
        multi: list[MultiTargetMarkdown] = []
        markdowns = [Markdown.load_file(t) for t in sources if t.suffix == ".md"]
        for md in markdowns:
            if not (md.path and md.front_matter and md.front_matter.documentation_of):
                continue

            if isinstance(md.front_matter.documentation_of, str):
                source_path = resolve_documentation_of(
                    md.front_matter.documentation_of,
                    basedir=md.path.parent,
                )
                if source_path in sources:
                    md.front_matter.documentation_of = source_path.as_posix()
                    single[source_path] = md
                else:
                    logger.warning(
                        "Markdown(%s) documentation_of: %s is not found.",
                        md.path,
                        md.front_matter.documentation_of,
                    )
            else:
                multi_documentation_of: list[pathlib.Path] = []
                for d in md.front_matter.documentation_of:
                    source_path = resolve_documentation_of(
                        d,
                        basedir=md.path.parent,
                    )
                    if source_path in sources:
                        multi_documentation_of.append(source_path)
                    else:
                        logger.warning(
                            "Markdown(%s) documentation_of: %s is not found.",
                            md.path,
                            d,
                        )
                if multi_documentation_of:
                    multi.append(
                        MultiTargetMarkdown(
                            path=md.path,
                            front_matter=md.front_matter,
                            content=md.content,
                            multi_documentation_of=multi_documentation_of,
                        )
                    )
                else:
                    logger.warning(
                        "Markdown(%s) documentation_of have no valid files.",
                        md.path,
                    )

        for m in multi:
            if m.front_matter and m.front_matter.keep_single:
                continue
            for source in m.multi_documentation_of:
                redirect_to = f"/{m.path.with_suffix('').as_posix()}"
                s = single.get(source)
                if s:
                    if s.front_matter:
                        s.front_matter.display = DocumentOutputMode.no_index
                        s.front_matter.redirect_to = redirect_to
                    else:
                        s.front_matter = FrontMatter(
                            redirect_to=redirect_to,
                            display=DocumentOutputMode.no_index,
                        )
                else:
                    s = Markdown(
                        content=b"",
                        front_matter=FrontMatter(
                            redirect_to=redirect_to,
                            display=DocumentOutputMode.no_index,
                        ),
                    )
                single[source] = s
        return UserMarkdowns(
            single=single,
            multi=multi,
        )


class _VerificationStatusFlag(enum.Flag):
    IS_LIBRARY = 0
    NOTHING = 0
    LIBRARY_NOTHING = IS_LIBRARY
    TEST_NOTHING = enum.auto()
    IS_TEST = TEST_NOTHING
    HAVE_AC = enum.auto()
    HAVE_WA = enum.auto()
    HAVE_SKIP = enum.auto()

    LIBRARY_AC_WA_SKIP = IS_LIBRARY | HAVE_AC | HAVE_WA | HAVE_SKIP
    LIBRARY_AC_WA = IS_LIBRARY | HAVE_AC | HAVE_WA
    LIBRARY_AC_SKIP = IS_LIBRARY | HAVE_AC | HAVE_SKIP
    LIBRARY_AC = IS_LIBRARY | HAVE_AC
    LIBRARY_WA_SKIP = IS_LIBRARY | HAVE_WA | HAVE_SKIP
    LIBRARY_WA = IS_LIBRARY | HAVE_WA
    LIBRARY_SKIP = IS_LIBRARY | HAVE_SKIP

    TEST_AC_WA_SKIP = IS_TEST | HAVE_AC | HAVE_WA | HAVE_SKIP
    TEST_AC_WA = IS_TEST | HAVE_AC | HAVE_WA
    TEST_AC_SKIP = IS_TEST | HAVE_AC | HAVE_SKIP
    TEST_AC = IS_TEST | HAVE_AC
    TEST_WA_SKIP = IS_TEST | HAVE_WA | HAVE_SKIP
    TEST_WA = IS_TEST | HAVE_WA
    TEST_SKIP = IS_TEST | HAVE_SKIP

    @classmethod
    @property
    def _conv_dict(cls) -> dict["_VerificationStatusFlag", StatusIcon]:
        try:
            d: dict["_VerificationStatusFlag", StatusIcon] = cls._conv_dict_attr
        except AttributeError:
            d = {
                cls.LIBRARY_AC_WA_SKIP: StatusIcon.LIBRARY_SOME_WA,
                cls.LIBRARY_AC_WA: StatusIcon.LIBRARY_SOME_WA,
                cls.LIBRARY_AC_SKIP: StatusIcon.LIBRARY_PARTIAL_AC,
                cls.LIBRARY_AC: StatusIcon.LIBRARY_ALL_AC,
                cls.LIBRARY_WA_SKIP: StatusIcon.LIBRARY_ALL_WA,
                cls.LIBRARY_WA: StatusIcon.LIBRARY_ALL_WA,
                cls.LIBRARY_SKIP: StatusIcon.LIBRARY_NO_TESTS,
                cls.LIBRARY_NOTHING: StatusIcon.LIBRARY_NO_TESTS,
                cls.TEST_AC_WA_SKIP: StatusIcon.TEST_WRONG_ANSWER,
                cls.TEST_AC_WA: StatusIcon.TEST_WRONG_ANSWER,
                cls.TEST_AC_SKIP: StatusIcon.TEST_WAITING_JUDGE,
                cls.TEST_AC: StatusIcon.TEST_ACCEPTED,
                cls.TEST_WA_SKIP: StatusIcon.TEST_WRONG_ANSWER,
                cls.TEST_WA: StatusIcon.TEST_WRONG_ANSWER,
                cls.TEST_SKIP: StatusIcon.TEST_WAITING_JUDGE,
                cls.TEST_NOTHING: StatusIcon.TEST_WAITING_JUDGE,
            }
            cls._conv_dict_attr = d
        return d

    @classmethod
    @property
    def _conv_dict_rev(cls) -> dict[StatusIcon, "_VerificationStatusFlag"]:
        try:
            d: dict[StatusIcon, "_VerificationStatusFlag"] = cls._conv_dict_rev_attr
        except AttributeError:
            d = {
                StatusIcon.LIBRARY_SOME_WA: cls.LIBRARY_AC_WA,
                StatusIcon.LIBRARY_PARTIAL_AC: cls.LIBRARY_AC_SKIP,
                StatusIcon.LIBRARY_ALL_AC: cls.LIBRARY_AC,
                StatusIcon.LIBRARY_ALL_WA: cls.LIBRARY_WA,
                StatusIcon.LIBRARY_NO_TESTS: cls.LIBRARY_NOTHING,
                StatusIcon.TEST_ACCEPTED: cls.TEST_AC,
                StatusIcon.TEST_WRONG_ANSWER: cls.TEST_WA,
                StatusIcon.TEST_WAITING_JUDGE: cls.TEST_NOTHING,
            }
            cls._conv_dict_rev_attr = d
        return d

    def to_status(self) -> StatusIcon:
        return self._conv_dict[self]

    @classmethod
    def from_status(cls, status: StatusIcon) -> "_VerificationStatusFlag":
        return cls._conv_dict_rev[status]


class SourceCodeStat(BaseModel):
    path: ForcePosixPath
    is_verification: bool
    verification_status: StatusIcon
    file_input: VerificationFile
    timestamp: datetime.datetime
    depends_on: SortedPathSet
    required_by: SortedPathSet
    verified_with: SortedPathSet
    verification_results: Optional[list[VerificationResult]] = None

    @staticmethod
    def resolve_dependency(
        *,
        input: VerificationInput,
        result: VerifyCommandResult,
        included_files: AbstractSet[pathlib.Path],
    ) -> dict[pathlib.Path, "SourceCodeStat"]:
        d: dict[pathlib.Path, SourceCodeStat] = {}
        statuses: dict[pathlib.Path, _VerificationStatusFlag] = {
            p: _VerificationStatusFlag.NOTHING for p in input.files.keys()
        }
        verification_results_dict: dict[pathlib.Path, list[VerificationResult]] = {}

        for p, r in result.files.items():
            if p not in included_files:
                continue
            st = _VerificationStatusFlag.NOTHING
            for v in r.verifications:
                if v.status == ResultStatus.SUCCESS:
                    st |= _VerificationStatusFlag.HAVE_AC
                elif v.status == ResultStatus.FAILURE:
                    st |= _VerificationStatusFlag.HAVE_WA
                elif v.status == ResultStatus.SKIPPED:
                    st |= _VerificationStatusFlag.HAVE_SKIP
            statuses[p] = st
            verification_results_dict[p] = r.verifications

        for group in input.scc():
            group &= included_files
            if not group:
                continue

            group_status = _VerificationStatusFlag.NOTHING

            for path in group:
                assert path in statuses
                group_status |= statuses[path]

            for path in group:
                depends_on = input.depends_on[path] & included_files
                required_by = input.required_by[path] & included_files
                verified_with = input.verified_with[path] & included_files

                for dep in depends_on:
                    statuses[dep] |= group_status

                timestamp = git.get_commit_time(input.transitive_depends_on[path])
                file_input = input.files[path]
                is_verification = file_input.is_verification()

                verification_results = verification_results_dict.get(path)

                assert not is_verification or verification_results is not None

                flag_status = group_status | (
                    _VerificationStatusFlag.IS_TEST
                    if is_verification
                    else _VerificationStatusFlag.IS_LIBRARY
                )

                d[path] = SourceCodeStat(
                    path=path,
                    file_input=file_input,
                    is_verification=is_verification,
                    depends_on=depends_on,
                    required_by=required_by,
                    verified_with=verified_with,
                    timestamp=timestamp,
                    verification_status=flag_status.to_status(),
                    verification_results=verification_results,
                )
        return d


class RenderJob(ABC):
    def render(self, dst: pathlib.Path):
        file = dst / self.destination_name
        file.mkdir(parents=True, exist_ok=True)
        with file.open("rb") as fp:
            self.write_to(fp)

    @abstractproperty
    def destination_name(self) -> pathlib.Path:
        ...

    @abstractmethod
    def write_to(self, fp: BinaryIO):
        ...

    @staticmethod
    def enumerate_jobs(
        *,
        sources: set[pathlib.Path],
        input: VerificationInput,
        result: VerifyCommandResult,
        config: ConfigYaml,
        index_md: Optional[Markdown] = None,
    ) -> list["RenderJob"]:
        def plain_content(source: pathlib.Path) -> Optional[RenderJob]:
            if source.suffix == ".md":
                md = Markdown.load_file(source)
                if md.front_matter and md.front_matter.documentation_of:
                    return None
            elif source.suffix == ".html":
                pass
            else:
                return None
            return PlainRenderJob(
                source_path=source,
                content=source.read_bytes(),
            )

        user_markdowns = UserMarkdowns.select_markdown(sources)

        logger.info(" %s source files...", len(sources))

        class SourceForDebug(BaseModel):
            sources: SortedPathSet
            markdowns: UserMarkdowns

        logger.debug(
            "source: %s",
            SourceForDebug(
                sources=sources,
                markdowns=user_markdowns,
            ),
        )
        with log.group("Resolve dependency"):
            stats_dict = SourceCodeStat.resolve_dependency(
                input=input,
                result=result,
                included_files=sources,
            )

        page_jobs: dict[pathlib.Path, PageRenderJob] = {}
        jobs: list[RenderJob] = []
        for source in sources:
            markdown = user_markdowns.single.get(source) or Markdown.make_default(
                source
            )
            stat = stats_dict.get(source)
            if not stat:
                plain_job = plain_content(source)
                if plain_job is not None:
                    jobs.append(plain_job)
                elif source.suffix != ".md":
                    logger.info("Skip file: %s", source.as_posix())
                continue
            group_dir = None
            if config.consolidate:
                consolidate = config.consolidate
                group_dir = next(
                    filter(lambda p: p in consolidate, source.parents), None
                )

            pj = PageRenderJob(
                source_path=source,
                group_dir=group_dir or source.parent,
                markdown=markdown,
                stat=stat,
                input=input,
                result=result,
                page_jobs=page_jobs,
            )

            if pj.display == DocumentOutputMode.never:
                continue

            page_jobs[pj.source_path] = pj
            jobs.append(pj)

        multis: list[MultiCodePageRenderJob] = []
        for md in user_markdowns.multi:
            group_dir = None
            if config.consolidate:
                consolidate = config.consolidate
                group_dir = next(
                    filter(lambda p: p in consolidate, md.path.parents), None
                )
            job = MultiCodePageRenderJob(
                markdown=md,
                group_dir=group_dir or md.path.parent,
                page_jobs=page_jobs,
            )

            multis.append(job)

            # if not md.front_matter.keep_single:
            #     for of in md.multi_documentation_of:
            #         pj = page_jobs.get(of)
            #         if pj:
            #             pj.render_link = job.to_render_link()

        jobs.extend(multis)
        jobs.append(
            IndexRenderJob(
                page_jobs=page_jobs,
                multicode_docs=multis,
                index_md=index_md,
            )
        )
        return jobs


@dataclass(frozen=True)
class PlainRenderJob(RenderJob):
    source_path: ForcePosixPath
    content: bytes

    @property
    def destination_name(self):
        return self.source_path

    def write_to(self, fp: BinaryIO):
        fp.write(self.content)


class MarkdownRenderJob(RenderJob):
    source_path: pathlib.Path
    markdown: Markdown

    @property
    def destination_name(self):
        return self.source_path

    def write_to(self, fp: BinaryIO):
        self.markdown.dump_merged(fp)


@dataclass(frozen=True)
class PageRenderJob(RenderJob):
    source_path: pathlib.Path
    group_dir: pathlib.Path
    markdown: Markdown
    stat: SourceCodeStat
    input: VerificationInput
    result: VerifyCommandResult
    page_jobs: dict[pathlib.Path, "PageRenderJob"]

    @property
    def is_verification(self):
        return self.stat.is_verification

    @property
    def display(self):
        return self.front_matter.display or DocumentOutputMode.visible

    def __str__(self) -> str:
        return f"PageRenderJob(source_path={repr(self.source_path)},markdown={repr(self.markdown)},stat={repr(self.stat)})"

    def validate_front_matter(self):
        front_matter = self.markdown.front_matter
        if front_matter and front_matter.documentation_of:
            if not isinstance(
                front_matter.documentation_of, str
            ) or self.source_path != pathlib.Path(front_matter.documentation_of):
                raise ValueError(
                    "PageRenderJob.path must equal front_matter.documentation_of."
                )

    def to_render_link(self, *, index: bool = False) -> Optional[RenderLink]:
        if (
            self.display == DocumentOutputMode.hidden
            or self.display == DocumentOutputMode.never
            or (index and self.display == DocumentOutputMode.no_index)
        ):
            return None
        return RenderLink(
            path=self.source_path,
            filename=self.source_path.relative_to(self.group_dir).as_posix(),
            title=self.front_matter.title,
            icon=self.stat.verification_status,
        )

    @cached_property
    def front_matter(self) -> FrontMatter:
        if self.markdown.front_matter:
            front_matter = self.markdown.front_matter.model_copy()
        else:
            front_matter = FrontMatter()
        front_matter.documentation_of = self.source_path.as_posix()
        if not front_matter.layout:
            front_matter.layout = "document"

        input_file = self.input.files.get(self.source_path)
        if not front_matter.title and (input_file and input_file.title):
            front_matter.title = input_file.title
        if not front_matter.display and (input_file and input_file.display):
            front_matter.display = input_file.display

        return front_matter

    @property
    def destination_name(self):
        return self.source_path.with_suffix(self.source_path.suffix + ".md")

    def write_to(self, fp: BinaryIO):
        self.validate_front_matter()
        front_matter = self.front_matter
        front_matter.data = self.get_page_data()
        Markdown(
            path=self.source_path,
            front_matter=front_matter,
            content=self.markdown.content,
        ).dump_merged(fp)

    def get_page_data(self) -> PageRenderData:
        depends_on = _paths_to_render_links(self.stat.depends_on, self.page_jobs)
        required_by = _paths_to_render_links(self.stat.required_by, self.page_jobs)
        verified_with = _paths_to_render_links(self.stat.verified_with, self.page_jobs)

        attributes = self.stat.file_input.document_attributes.copy()
        problem = next(
            (
                v.problem
                for v in self.stat.file_input.verification_list
                if isinstance(v, ProblemVerification)
            ),
            None,
        )
        if problem:
            attributes.setdefault("PROBLEM", problem)

        code = read_text_normalized(self.source_path)

        embedded = [EmbeddedCode(name="default", code=code)]
        for s in self.stat.file_input.additonal_sources:
            embedded.append(
                EmbeddedCode(name=s.name, code=read_text_normalized(s.path))
            )

        return PageRenderData(
            path=self.source_path,
            path_extension=self.source_path.suffix.lstrip("."),
            title=self.front_matter.title,
            embedded=embedded,
            timestamp=self.stat.timestamp,
            attributes=attributes,
            testcases=[
                EnvTestcaseResult(
                    name=c.name,
                    status=c.status,
                    elapsed=c.elapsed,
                    memory=c.memory,
                    environment=v.verification_name,
                )
                for v in self.stat.verification_results
                for c in (v.testcases or [])
            ]
            if self.stat.verification_results
            else None,
            verification_status=self.stat.verification_status,
            is_verification_file=self.stat.is_verification,
            is_failed=self.stat.verification_status.is_failed,
            document_path=self.markdown.path,
            dependencies=[
                Dependency(type="Depends on", files=depends_on),
                Dependency(type="Required by", files=required_by),
                Dependency(type="Verified with", files=verified_with),
            ],
            depends_on=[link.path for link in depends_on],
            required_by=[link.path for link in required_by],
            verified_with=[link.path for link in verified_with],
        )


@dataclass(frozen=True)
class MultiCodePageRenderJob(RenderJob):
    markdown: MultiTargetMarkdown
    group_dir: pathlib.Path
    page_jobs: dict[pathlib.Path, "PageRenderJob"]

    def __str__(self) -> str:
        return f"MultiCodePageRenderJob(multi_documentation_of={repr(self.markdown.multi_documentation_of)})"

    @cached_property
    def jobs(self) -> list[PageRenderJob]:
        jobs: list[PageRenderJob] = []
        for m in self.markdown.multi_documentation_of:
            job = self.page_jobs.get(m)
            if not job:
                continue
            jobs.append(job)
        return jobs

    @cached_property
    def verification_status(self) -> StatusIcon:
        flag = _VerificationStatusFlag.NOTHING
        for job in self.jobs:
            flag |= _VerificationStatusFlag.from_status(job.stat.verification_status)
        return flag.to_status()

    @property
    def is_verification(self):
        return self.verification_status.is_test

    @property
    def display(self):
        return self.markdown.front_matter.display or DocumentOutputMode.visible

    @property
    def destination_name(self) -> pathlib.Path:
        return self.markdown.path

    def to_render_link(self, *, index: bool = False) -> RenderLink:
        return RenderLink(
            path=self.markdown.path.with_suffix(""),
            filename=self.markdown.path.relative_to(self.group_dir).as_posix(),
            title=self.markdown.front_matter.title,
            icon=self.verification_status,
        )

    def write_to(self, fp: BinaryIO):
        front_matter = self.markdown.front_matter
        front_matter.layout = "multidoc"
        front_matter.data = self.get_page_data()
        Markdown(
            path=self.markdown.path,
            front_matter=front_matter,
            content=self.markdown.content,
        ).dump_merged(fp)

    def get_page_data(self) -> MultiCodePageData:
        codes = [
            CodePageData.model_validate(
                {"document_content": normalize_bytes_text(j.markdown.content)}
                | j.get_page_data().model_dump(),
            )
            for j in self.jobs
        ]

        multi_documentation_of_set = set(self.markdown.multi_documentation_of)
        multi_documentation_of_set.add(self.markdown.path)
        depends_on_paths = (
            set(chain.from_iterable(j.stat.depends_on for j in self.jobs))
            - multi_documentation_of_set
        )
        required_by_paths = (
            set(chain.from_iterable(j.stat.required_by for j in self.jobs))
            - multi_documentation_of_set
        )
        verified_with_paths = (
            set(chain.from_iterable(j.stat.verified_with for j in self.jobs))
            - multi_documentation_of_set
        )

        depends_on = _paths_to_render_links(depends_on_paths, self.page_jobs)
        required_by = _paths_to_render_links(required_by_paths, self.page_jobs)
        verified_with = _paths_to_render_links(verified_with_paths, self.page_jobs)

        return MultiCodePageData(
            path=self.markdown.path,
            verification_status=self.verification_status,
            is_failed=any(c.is_failed for c in codes),
            codes=codes,
            dependencies=[
                Dependency(type="Depends on", files=depends_on),
                Dependency(type="Required by", files=required_by),
                Dependency(type="Verified with", files=verified_with),
            ],
        )


@dataclass
class IndexRenderJob(RenderJob):
    page_jobs: dict[pathlib.Path, "PageRenderJob"]
    multicode_docs: list[MultiCodePageRenderJob]
    index_md: Optional[Markdown] = None

    def __str__(self) -> str:
        @dataclass
        class _IndexRenderJob:
            job_paths: Iterable[pathlib.Path]

        s = repr(
            _IndexRenderJob(
                job_paths=self.page_jobs.keys(),
            )
        )
        index = s.find("_IndexRenderJob")
        return s[index + 1 :]

    @property
    def destination_name(self):
        return pathlib.Path("index.md")

    def write_to(self, fp: BinaryIO):
        Markdown(
            path=self.destination_name,
            front_matter=FrontMatter(
                layout="toppage",
                data=self.get_page_data(),
            ),
            content=self.index_md.content if self.index_md else b"",
        ).dump_merged(fp)

    def get_page_data(self) -> IndexRenderData:
        library_categories: dict[str, list[RenderLink]] = {}
        verification_categories: dict[str, list[RenderLink]] = {}
        for job in chain.from_iterable([self.page_jobs.values(), self.multicode_docs]):
            if job.display != DocumentOutputMode.visible:
                continue
            if job.is_verification:
                categories = verification_categories
            else:
                categories = library_categories

            directory = job.group_dir
            category = directory.as_posix()
            if category == ".":
                category = ""
            elif not category.endswith("/"):
                category = f"{category}/"

            if category not in categories:
                categories[category] = []

            link = job.to_render_link(index=True)
            if link:
                categories[category].append(link)

        def _build_categories_list(
            categories: dict[str, list[RenderLink]]
        ) -> list[CategorizedIndex]:
            return sorted(
                (
                    CategorizedIndex(
                        name=category,
                        pages=sorted(pages, key=lambda p: p.path.as_posix()),
                    )
                    for category, pages in categories.items()
                ),
                key=lambda d: d.name,
            )

        return IndexRenderData(
            top=[
                IndexFiles(
                    type="Library Files",
                    categories=_build_categories_list(library_categories),
                ),
                IndexFiles(
                    type="Verification Files",
                    categories=_build_categories_list(verification_categories),
                ),
            ],
        )



import pathlib
from typing import Annotated, Iterable

from pydantic.functional_serializers import PlainSerializer

ForcePosixPath = Annotated[
    pathlib.Path,
    PlainSerializer(lambda x: x.as_posix(), return_type=str, when_used="json"),
]


def _sorted(items: Iterable[ForcePosixPath]):
    return sorted(items, key=lambda x: str.casefold(x.as_posix()))


SortedPathSet = Annotated[
    set[ForcePosixPath],
    PlainSerializer(_sorted, return_type=list[ForcePosixPath], when_used="json"),
]


SortedPathList = Annotated[
    list[ForcePosixPath],
    PlainSerializer(_sorted, return_type=list[ForcePosixPath]),
]

RelativeDirectoryPath = Annotated[
    pathlib.Path,
    PlainSerializer(
        lambda x: f"{x.as_posix()}/" if x != pathlib.Path(".") else "",
        return_type=str,
        when_used="json",
    ),
]



# flake8: noqa E501
import json
from pathlib import Path

from competitive_verifier.models import VerificationInput

test_input = VerificationInput.model_validate(
    {
        "files": {
            "foo/bar1.py": {},
            "foo/bar2.py": {"dependencies": ["foo/bar1.py"]},
            "foo/baz.py": {},
            "foo/barbaz.py": {
                "dependencies": [
                    "foo/bar2.py",
                    "foo/baz.py",
                ],
            },
            "hoge/1.py": {},
            "hoge/hoge.py": {
                "dependencies": [
                    "hoge/fuga.py",
                    "hoge/1.py",
                ],
            },
            "hoge/piyo.py": {
                "dependencies": [
                    "hoge/fuga.py",
                    "hoge/hoge.py",
                ],
            },
            "hoge/fuga.py": {"dependencies": ["hoge/piyo.py"]},
            "hoge/piyopiyo.py": {"dependencies": ["hoge/piyo.py"]},
            "test/test.py": {
                "verification": [{"status": "success", "type": "const"}],
                "dependencies": ["hoge/piyopiyo.py"],
            },
        }
    }
)


def test_to_json():
    assert (
        VerificationInput.model_validate_json(
            test_input.model_dump_json(exclude_none=True)
        )
        == test_input
    )

    obj = VerificationInput.model_validate(
        {
            "files": {
                "foo/bar.py": {},
                "foo/baz.py": {
                    "path": "foo/baz.py",
                    "document_attributes": {"title": "foo-baz"},
                    "dependencies": ["foo/bar.py"],
                    "verification": [
                        {
                            "type": "const",
                            "status": "success",
                        }
                    ],
                    "additonal_sources": [{"name": "dummy", "path": "tmp/dumm.py"}],
                },
            },
        }
    )
    assert json.loads(obj.model_dump_json(exclude_none=True)) == {
        "files": {
            "foo/bar.py": {
                "dependencies": [],
                "document_attributes": {},
                "verification": [],
                "additonal_sources": [],
            },
            "foo/baz.py": {
                "document_attributes": {"title": "foo-baz"},
                "dependencies": ["foo/bar.py"],
                "verification": [
                    {
                        "type": "const",
                        "status": "success",
                    }
                ],
                "additonal_sources": [{"name": "dummy", "path": "tmp/dumm.py"}],
            },
        },
    }


def test_repr():
    obj = VerificationInput.model_validate(
        {
            "files": {
                "foo/bar.py": {},
                "foo/baz.py": {
                    "path": "foo/baz.py",
                    "document_attributes": {
                        "title": "foo-baz",
                    },
                    "dependencies": ["foo/bar.py"],
                    "verification": [
                        {
                            "type": "const",
                            "status": "success",
                        }
                    ],
                    "additonal_sources": [{"name": "dummy", "path": "tmp/dumm.py"}],
                },
            },
        }
    )
    print(repr(obj))
    assert repr(obj) == (
        "VerificationInput("
        f"files={{{repr(Path('foo/bar.py'))}: "
        f"VerificationFile(dependencies=set(), verification=[], document_attributes={{}}, additonal_sources=[]),"
        f" {repr(Path('foo/baz.py'))}: "
        f"VerificationFile(dependencies={{{repr(Path('foo/bar.py'))}}}, verification=[ConstVerification(name=None, type='const', status=<ResultStatus.SUCCESS: 'success'>)], document_attributes={{'title': {repr('foo-baz')}}}, additonal_sources=[AddtionalSource(name='dummy', path={repr(Path('tmp/dumm.py'))})])"
        f"}})"
    )


def test_transitive_depends_on():
    simple = {
        "foo/bar1.py": ["foo/bar1.py"],
        "foo/bar2.py": ["foo/bar1.py", "foo/bar2.py"],
        "foo/baz.py": ["foo/baz.py"],
        "foo/barbaz.py": ["foo/barbaz.py", "foo/baz.py", "foo/bar1.py", "foo/bar2.py"],
        "hoge/1.py": ["hoge/1.py"],
        "hoge/hoge.py": ["hoge/hoge.py", "hoge/piyo.py", "hoge/fuga.py", "hoge/1.py"],
        "hoge/piyo.py": ["hoge/hoge.py", "hoge/piyo.py", "hoge/fuga.py", "hoge/1.py"],
        "hoge/fuga.py": ["hoge/hoge.py", "hoge/piyo.py", "hoge/fuga.py", "hoge/1.py"],
        "hoge/piyopiyo.py": [
            "hoge/piyopiyo.py",
            "hoge/hoge.py",
            "hoge/piyo.py",
            "hoge/fuga.py",
            "hoge/1.py",
        ],
        "test/test.py": [
            "hoge/1.py",
            "hoge/fuga.py",
            "hoge/hoge.py",
            "hoge/piyo.py",
            "hoge/piyopiyo.py",
            "test/test.py",
        ],
    }
    expected = {Path(p): set(Path(s) for s in d) for p, d in simple.items()}

    assert test_input.transitive_depends_on == expected
    assert test_input.transitive_depends_on is test_input.transitive_depends_on
    assert test_input.transitive_depends_on == expected


def test_depends_on():
    simple = {
        "foo/bar1.py": [],
        "foo/bar2.py": [("foo/bar1.py")],
        "foo/baz.py": [],
        "foo/barbaz.py": ["foo/bar2.py", "foo/baz.py"],
        "hoge/1.py": [],
        "hoge/hoge.py": ["hoge/fuga.py", "hoge/1.py"],
        "hoge/piyo.py": ["hoge/fuga.py", "hoge/hoge.py"],
        "hoge/fuga.py": ["hoge/piyo.py"],
        "hoge/piyopiyo.py": ["hoge/piyo.py"],
        "test/test.py": ["hoge/piyopiyo.py"],
    }
    expected = {Path(p): set(Path(s) for s in d) for p, d in simple.items()}

    assert test_input.depends_on == expected
    assert test_input.depends_on is test_input.depends_on
    assert test_input.depends_on == expected


def test_required_by():
    simple = {
        "foo/bar1.py": ["foo/bar2.py"],
        "foo/bar2.py": ["foo/barbaz.py"],
        "foo/baz.py": ["foo/barbaz.py"],
        "foo/barbaz.py": [],
        "hoge/1.py": ["hoge/hoge.py"],
        "hoge/hoge.py": ["hoge/piyo.py"],
        "hoge/piyo.py": ["hoge/fuga.py", "hoge/piyopiyo.py"],
        "hoge/fuga.py": ["hoge/piyo.py", "hoge/hoge.py"],
        "hoge/piyopiyo.py": [],
        "test/test.py": [],
    }
    expected = {Path(p): set(Path(s) for s in d) for p, d in simple.items()}

    assert test_input.required_by == expected
    assert test_input.required_by is test_input.required_by
    assert test_input.required_by == expected


def test_verified_with():
    simple = {
        "foo/bar1.py": [],
        "foo/bar2.py": [],
        "foo/baz.py": [],
        "foo/barbaz.py": [],
        "hoge/1.py": [],
        "hoge/hoge.py": [],
        "hoge/piyo.py": [],
        "hoge/fuga.py": [],
        "hoge/piyopiyo.py": ["test/test.py"],
        "test/test.py": [],
    }
    expected = {Path(p): set(Path(s) for s in d) for p, d in simple.items()}
    assert test_input.verified_with == expected
    assert test_input.verified_with is test_input.verified_with
    assert test_input.verified_with == expected



import sys
from logging import getLogger
from typing import BinaryIO

from pydantic import BaseModel, ConfigDict, Field

from competitive_verifier.oj.verify.languages.cplusplus import (
    CPlusPlusLanguage,
    OjVerifyCPlusPlusConfig,
)
from competitive_verifier.oj.verify.languages.go import GoLanguage, OjVerifyGoConfig
from competitive_verifier.oj.verify.languages.haskell import (
    HaskellLanguage,
    OjVerifyHaskellConfig,
)
from competitive_verifier.oj.verify.languages.java import (
    JavaLanguage,
    OjVerifyJavaConfig,
)
from competitive_verifier.oj.verify.languages.nim import NimLanguage, OjVerifyNimConfig
from competitive_verifier.oj.verify.languages.python import PythonLanguage
from competitive_verifier.oj.verify.languages.ruby import (
    OjVerifyRubyConfig,
    RubyLanguage,
)
from competitive_verifier.oj.verify.languages.rust import (
    OjVerifyRustConfig,
    RustLanguage,
)
from competitive_verifier.oj.verify.languages.user_defined import UserDefinedLanguage
from competitive_verifier.oj.verify.models import Language, OjVerifyUserDefinedConfig

if sys.version_info >= (3, 11):
    import tomllib as tomli
else:
    import tomli

logger = getLogger(__name__)


class OjVerifyLanguageConfigDict(BaseModel):
    model_config = ConfigDict(extra="allow")
    __pydantic_extra__: dict[str, OjVerifyUserDefinedConfig]  # pyright: ignore

    cpp: OjVerifyCPlusPlusConfig = Field(default_factory=OjVerifyCPlusPlusConfig)
    go: OjVerifyGoConfig = Field(default_factory=OjVerifyGoConfig)
    haskell: OjVerifyHaskellConfig = Field(default_factory=OjVerifyHaskellConfig)
    java: OjVerifyJavaConfig = Field(default_factory=OjVerifyJavaConfig)
    nim: OjVerifyNimConfig = Field(default_factory=OjVerifyNimConfig)
    ruby: OjVerifyRubyConfig = Field(default_factory=OjVerifyRubyConfig)
    rust: OjVerifyRustConfig = Field(default_factory=OjVerifyRustConfig)


class OjVerifyConfig(BaseModel):
    languages: OjVerifyLanguageConfigDict = Field(
        default_factory=OjVerifyLanguageConfigDict  # pyright: ignore
    )

    @classmethod
    def load(cls, fp: BinaryIO) -> "OjVerifyConfig":
        return OjVerifyConfig.model_validate(tomli.load(fp))

    def get_dict(self) -> dict[str, Language]:
        languages = self.languages
        d: dict[str, Language] = {}
        d[".cpp"] = CPlusPlusLanguage(config=languages.cpp)
        d[".hpp"] = d[".cpp"]
        d[".cc"] = d[".cpp"]
        d[".h"] = d[".cpp"]
        d[".nim"] = NimLanguage(config=languages.nim)
        d[".py"] = PythonLanguage()
        d[".hs"] = HaskellLanguage(config=languages.haskell)
        d[".ruby"] = RubyLanguage(config=languages.ruby)
        d[".go"] = GoLanguage(config=languages.go)
        d[".java"] = JavaLanguage(config=languages.java)
        d[".rs"] = RustLanguage(config=languages.rust)

        for ext, lang_config in languages.__pydantic_extra__.items():
            d["." + ext] = UserDefinedLanguage(extension=ext, config=lang_config)

        return d



import json
import pathlib
from typing import Optional, Protocol

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.oj.verify.languages import special_comments
from competitive_verifier.oj_resolve.main import main

from .data.integration_data import IntegrationData
from .types import FilePaths


@pytest.fixture
def setenv_resolve(mocker: MockerFixture):
    special_comments.list_special_comments.cache_clear()
    special_comments.list_embedded_urls.cache_clear()

    def python_get_execute_command(
        path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        return f"env PYTHONPATH={basedir.resolve().as_posix()} python {path.as_posix()}"

    mocker.patch(
        "competitive_verifier.oj.verify.languages.python.PythonLanguageEnvironment.get_execute_command",
        side_effect=python_get_execute_command,
    )


class _ArgsFunc(Protocol):
    def __call__(
        self,
        *,
        bundle: bool = True,
        include: Optional[list[str]] = None,
        exclude: Optional[list[str]] = None,
        config: Optional[str] = None,
    ) -> list[str]:
        ...


@pytest.fixture
def make_args() -> _ArgsFunc:
    def _make_args(
        *,
        bundle: bool = False,
        include: Optional[list[str]] = None,
        exclude: Optional[list[str]] = None,
        config: Optional[str] = None,
    ) -> list[str]:
        args: list[str] = []
        if not bundle:
            args.append("--no-bundle")
        if include is not None:
            args.append("--include")
            args.extend(include)
        if exclude is not None:
            args.append("--exclude")
            args.extend(exclude)
        if config is not None:
            args.append("--config")
            args.append(config)
        return args

    return _make_args


@pytest.mark.usefixtures("additional_path")
@pytest.mark.order(-1000)
class TestCommandOjResolve:
    @pytest.mark.each_language_integration
    @pytest.mark.integration
    @pytest.mark.usefixtures("setenv_resolve")
    def test_oj_resolve_by_lang_data(
        self,
        make_args: _ArgsFunc,
        integration_data: IntegrationData,
        capfd: pytest.CaptureFixture[str],
    ):
        expected = integration_data.expected_verify_json()
        args = make_args(
            include=integration_data.include_path,
            exclude=integration_data.exclude_path,
            config=integration_data.config_path,
            bundle=True,
        )
        main(args)

        stdout = capfd.readouterr().out
        resolved = json.loads(stdout)
        assert resolved == expected

        verify = integration_data.config_dir_path / "verify.json"
        verify.parent.mkdir(parents=True, exist_ok=True)
        verify.write_text(stdout, encoding="utf-8")

    @pytest.mark.usefixtures("setenv_resolve")
    def test_without_include_exclude(
        self,
        make_args: _ArgsFunc,
        monkeypatch: pytest.MonkeyPatch,
        file_paths: FilePaths,
        capfd: pytest.CaptureFixture[str],
    ):
        monkeypatch.chdir(file_paths.root / "IncludeExclude")
        args = make_args(
            config="config.toml",
            bundle=True,
        )
        main(args)

        stdout = capfd.readouterr().out
        resolved = json.loads(stdout)
        assert resolved == {
            "files": {
                "a1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["a1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "a2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["a2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "b1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["b1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "b2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["b2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "c1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["c1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "c2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["c2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/a1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/a1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/a2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/a2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/b1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/b1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/b2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/b2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/c1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/c1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/c2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/c2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
            },
        }

    @pytest.mark.usefixtures("setenv_resolve")
    def test_with_include1(
        self,
        make_args: _ArgsFunc,
        monkeypatch: pytest.MonkeyPatch,
        file_paths: FilePaths,
        capfd: pytest.CaptureFixture[str],
    ):
        monkeypatch.chdir(file_paths.root / "IncludeExclude")
        args = make_args(
            include=["subdir/"],
            config="config.toml",
            bundle=True,
        )
        main(args)

        stdout = capfd.readouterr().out
        resolved = json.loads(stdout)
        assert resolved == {
            "files": {
                "subdir/a1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/a1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/a2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/a2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/b1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/b1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/b2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/b2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/c1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/c1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/c2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/c2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
            },
        }

    @pytest.mark.usefixtures("setenv_resolve")
    def test_with_include2(
        self,
        make_args: _ArgsFunc,
        monkeypatch: pytest.MonkeyPatch,
        file_paths: FilePaths,
        capfd: pytest.CaptureFixture[str],
    ):
        monkeypatch.chdir(file_paths.root / "IncludeExclude")
        args = make_args(
            include=["**a*.txt"],
            config="config.toml",
            bundle=True,
        )
        main(args)

        stdout = capfd.readouterr().out
        resolved = json.loads(stdout)
        assert resolved == {
            "files": {
                "a1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["a1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "a2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["a2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/a1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/a1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/a2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/a2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
            },
        }

    @pytest.mark.usefixtures("setenv_resolve")
    def test_with_exclude1(
        self,
        make_args: _ArgsFunc,
        monkeypatch: pytest.MonkeyPatch,
        file_paths: FilePaths,
        capfd: pytest.CaptureFixture[str],
    ):
        monkeypatch.chdir(file_paths.root / "IncludeExclude")
        args = make_args(
            exclude=["subdir/"],
            config="config.toml",
            bundle=True,
        )
        main(args)

        stdout = capfd.readouterr().out
        resolved = json.loads(stdout)
        assert resolved == {
            "files": {
                "a1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["a1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "a2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["a2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "b1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["b1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "b2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["b2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "c1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["c1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "c2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["c2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
            },
        }

    @pytest.mark.usefixtures("setenv_resolve")
    def test_with_exclude2(
        self,
        make_args: _ArgsFunc,
        monkeypatch: pytest.MonkeyPatch,
        file_paths: FilePaths,
        capfd: pytest.CaptureFixture[str],
    ):
        monkeypatch.chdir(file_paths.root / "IncludeExclude")
        args = make_args(
            exclude=["a*.txt", "b*.txt"],
            config="config.toml",
            bundle=True,
        )
        main(args)

        stdout = capfd.readouterr().out
        resolved = json.loads(stdout)
        assert resolved == {
            "files": {
                "c1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["c1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "c2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["c2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/a1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/a1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/a2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/a2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/b1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/b1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/b2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/b2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/c1.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/c1.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
                "subdir/c2.txt": {
                    "additonal_sources": [],
                    "dependencies": ["subdir/c2.txt"],
                    "document_attributes": {},
                    "verification": [],
                },
            },
        }



import argparse
import logging
import pathlib
import sys
from logging import getLogger
from typing import Optional

import competitive_verifier.config as config
import competitive_verifier.merge_result.main as merge_result
from competitive_verifier.arg import (
    add_ignore_error_argument,
    add_include_exclude_argument,
    add_result_json_argument,
    add_verbose_argument,
    add_verify_files_json_argument,
    add_write_summary_argument,
)
from competitive_verifier.log import configure_stderr_logging
from competitive_verifier.models import VerificationInput, VerifyCommandResult

from .. import config as conf
from .builder import DocumentBuilder

logger = getLogger(__name__)


def get_default_docs_dir() -> pathlib.Path:
    default_docs_dir = conf.get_config_dir() / "docs"
    oj_verify_docs_dir = pathlib.Path(".verify-helper/docs")
    if not default_docs_dir.exists() and oj_verify_docs_dir.exists():
        return oj_verify_docs_dir
    return default_docs_dir


def run_impl(
    input: VerificationInput,
    result: VerifyCommandResult,
    *,
    destination_dir: pathlib.Path,
    docs_dir: Optional[pathlib.Path],
    include: Optional[list[str]],
    exclude: Optional[list[str]],
) -> bool:
    logger.debug("input=%s", input.model_dump_json(exclude_none=True))
    logger.debug("result=%s", result.model_dump_json(exclude_none=True))
    builder = DocumentBuilder(
        input=input,
        result=result,
        docs_dir=docs_dir or get_default_docs_dir(),
        destination_dir=destination_dir,
        include=include,
        exclude=exclude,
    )
    return builder.build()


def run(args: argparse.Namespace) -> bool:
    default_level = logging.INFO
    if args.verbose:
        default_level = logging.DEBUG
    configure_stderr_logging(default_level)

    logger.debug("arguments=%s", vars(args))
    logger.info("verify_files_json=%s", str(args.verify_files_json))
    logger.info("result_json=%s", [str(p) for p in args.result_json])

    input = VerificationInput.parse_file_relative(args.verify_files_json)
    result = merge_result.run_impl(
        *args.result_json,
        write_summary=args.write_summary,
    )
    return (
        run_impl(
            input,
            result,
            docs_dir=args.docs,
            destination_dir=args.destination,
            include=args.include,
            exclude=args.exclude,
        )
        or args.ignore_error
    )


def argument(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    add_verbose_argument(parser)
    add_verify_files_json_argument(parser)
    add_result_json_argument(parser)
    add_ignore_error_argument(parser)
    add_write_summary_argument(parser)
    add_include_exclude_argument(parser)

    destination = config.get_config_dir() / "_jekyll"
    parser.add_argument(
        "--docs",
        type=pathlib.Path,
        help=f"Document settings directory. default: {get_default_docs_dir().as_posix()}",
    )
    parser.add_argument(
        "--destination",
        type=pathlib.Path,
        default=destination,
        help=f"Output directory for markdown document. default: {destination.as_posix()}",
    )
    return parser


def main(args: Optional[list[str]] = None) -> None:
    try:
        parsed = argument(argparse.ArgumentParser()).parse_args(args)
        if not run(parsed):
            sys.exit(1)
    except Exception as e:
        sys.stderr.write(str(e))
        sys.exit(2)



import os
import pathlib

import onlinejudge._implementation.utils
import onlinejudge.utils

COMPETITIVE_VERIFY_CONFIG_PATH = "COMPETITIVE_VERIFY_CONFIG_PATH"


def get_config_dir():
    p = pathlib.Path(os.getenv(COMPETITIVE_VERIFY_CONFIG_PATH, ".competitive-verifier"))
    onlinejudge._implementation.utils.user_cache_dir = _get_cache_dir(p)
    return p


def _get_cache_dir(config_dir: pathlib.Path):
    return config_dir / "cache"


def get_cache_dir():
    return _get_cache_dir(get_config_dir())



import shutil
from typing import Any

import pytest

from ..types import ConfigDirSetter, FilePaths
from .integration_data import IntegrationData


class RustWithoutConfigData(IntegrationData):
    def __init__(
        self,
        monkeypatch: pytest.MonkeyPatch,
        set_config_dir: ConfigDirSetter,
        file_paths: FilePaths,
    ) -> None:
        super().__init__(monkeypatch, set_config_dir, file_paths)

    def check_envinronment(self) -> bool:
        return bool(shutil.which("rustc"))

    @classmethod
    @property
    def input_name(cls) -> str:
        return "RustData"

    def expected_verify_json(self) -> dict[str, Any]:
        return dict(
            {
                "files": {
                    "crates/helloworld/hello/src/lib.rs": {
                        "additonal_sources": [],
                        "dependencies": ["crates/helloworld/hello/src/lib.rs"],
                        "document_attributes": {"links": []},
                        "verification": [],
                    },
                    "crates/helloworld/world/src/lib.rs": {
                        "additonal_sources": [],
                        "dependencies": ["crates/helloworld/world/src/lib.rs"],
                        "document_attributes": {"links": []},
                        "verification": [],
                    },
                    "crates/io/input/src/lib.rs": {
                        "additonal_sources": [],
                        "dependencies": [
                            "crates/io/input/src/lib.rs",
                            "crates/io/scanner/src/lib.rs",
                        ],
                        "document_attributes": {"links": []},
                        "verification": [],
                    },
                    "crates/io/scanner/src/lib.rs": {
                        "additonal_sources": [],
                        "dependencies": ["crates/io/scanner/src/lib.rs"],
                        "document_attributes": {"links": []},
                        "verification": [],
                    },
                    "src/lib.rs": {
                        "additonal_sources": [],
                        "dependencies": [
                            "crates/helloworld/hello/src/lib.rs",
                            "crates/helloworld/world/src/lib.rs",
                            "crates/io/input/src/lib.rs",
                            "crates/io/scanner/src/lib.rs",
                            "src/lib.rs",
                        ],
                        "document_attributes": {"links": []},
                        "verification": [],
                    },
                    "verification/src/bin/aizu-online-judge-itp1-1-a.rs": {
                        "additonal_sources": [],
                        "dependencies": [
                            "crates/helloworld/hello/src/lib.rs",
                            "crates/helloworld/world/src/lib.rs",
                            "crates/io/input/src/lib.rs",
                            "verification/src/bin/aizu-online-judge-itp1-1-a.rs",
                        ],
                        "document_attributes": {
                            "PROBLEM": "https://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=ITP1_1_A",
                            "links": [
                                "https://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=ITP1_1_A"
                            ],
                        },
                        "verification": [
                            {
                                "command": str(
                                    self.targets_path
                                    / "target/release/aizu-online-judge-itp1-1-a"
                                ),
                                "compile": "cd "
                                f"{self.targets_path/'verification/src/bin'} "
                                "&& "
                                "cargo "
                                "build "
                                "--release "
                                "--bin "
                                "aizu-online-judge-itp1-1-a",
                                "name": "Rust",
                                "problem": "https://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=ITP1_1_A",
                                "type": "problem",
                            }
                        ],
                    },
                    "verification/src/bin/library-checker-aplusb.rs": {
                        "additonal_sources": [],
                        "dependencies": [
                            "crates/helloworld/hello/src/lib.rs",
                            "crates/helloworld/world/src/lib.rs",
                            "crates/io/input/src/lib.rs",
                            "verification/src/bin/library-checker-aplusb.rs",
                        ],
                        "document_attributes": {
                            "PROBLEM": "https://judge.yosupo.jp/problem/aplusb",
                            "links": ["https://judge.yosupo.jp/problem/aplusb"],
                        },
                        "verification": [
                            {
                                "command": str(
                                    self.targets_path
                                    / "target/release/library-checker-aplusb"
                                ),
                                "compile": "cd "
                                f"{self.targets_path/'verification/src/bin'} "
                                "&& "
                                "cargo "
                                "build "
                                "--release "
                                "--bin "
                                "library-checker-aplusb",
                                "name": "Rust",
                                "problem": "https://judge.yosupo.jp/problem/aplusb",
                                "type": "problem",
                            }
                        ],
                    },
                },
            }
        )

    def expected_verify_result(self) -> dict[str, Any]:
        return {
            "files": {
                "verification/src/bin/aizu-online-judge-itp1-1-a.rs": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 3414.0,
                            "heaviest": 369.0,
                            "last_execution_time": "2046-08-15T20:50:50.570000-05:00",
                            "slowest": 341.0,
                            "status": "success",
                            "testcases": [
                                {
                                    "elapsed": 7.94,
                                    "memory": 5.09,
                                    "name": "judge_data",
                                    "status": "AC",
                                }
                            ],
                            "verification_name": "Rust",
                        }
                    ],
                },
                "verification/src/bin/library-checker-aplusb.rs": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 9167.0,
                            "heaviest": 758.0,
                            "last_execution_time": "2003-08-01T18:18:37.670000+05:00",
                            "slowest": 916.0,
                            "status": "success",
                            "testcases": [
                                {
                                    "elapsed": 4.35,
                                    "memory": 2.83,
                                    "name": "example_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 7.5,
                                    "memory": 35.22,
                                    "name": "example_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 3.53,
                                    "memory": 71.7,
                                    "name": "random_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.82,
                                    "memory": 81.21,
                                    "name": "random_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 0.12,
                                    "memory": 66.34,
                                    "name": "random_02",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 2.58,
                                    "memory": 38.37,
                                    "name": "random_03",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 2.95,
                                    "memory": 29.66,
                                    "name": "random_04",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 2.71,
                                    "memory": 59.01,
                                    "name": "random_05",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 7.2,
                                    "memory": 97.93,
                                    "name": "random_06",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.44,
                                    "memory": 72.71,
                                    "name": "random_07",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 3.02,
                                    "memory": 78.97,
                                    "name": "random_08",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 1.52,
                                    "memory": 18.84,
                                    "name": "random_09",
                                    "status": "AC",
                                },
                            ],
                            "verification_name": "Rust",
                        }
                    ],
                },
            },
            "total_seconds": 2547.12,
        }



from typing import Iterable, Union

import pytest

from competitive_verifier.download.main import UrlOrVerificationFile, parse_urls
from competitive_verifier.models import ProblemVerification, VerificationFile


def get_problem_command(url: str) -> ProblemVerification:
    return ProblemVerification(command="true", problem=url)


_SomeUrlOrVerificationFile = Union[
    UrlOrVerificationFile, Iterable[UrlOrVerificationFile]
]
test_parse_urls_params: list[tuple[_SomeUrlOrVerificationFile, set[str]]] = [
    (
        "http://example.com",
        {
            "http://example.com",
        },
    ),
    (
        VerificationFile(
            verification=[
                get_problem_command("http://example.com/alpha"),
                get_problem_command("http://example.com/beta"),
                get_problem_command("http://example.com/gamma"),
                get_problem_command("http://example.com/delta"),
            ]
        ),
        {
            "http://example.com/alpha",
            "http://example.com/beta",
            "http://example.com/gamma",
            "http://example.com/delta",
        },
    ),
    (
        [
            VerificationFile(
                verification=[
                    get_problem_command("http://example.com/alpha"),
                    get_problem_command("http://example.com/beta"),
                    get_problem_command("http://example.com/gamma"),
                    get_problem_command("http://example.com/delta"),
                ]
            ),
            VerificationFile(
                verification=[
                    get_problem_command("https://example.com/alpha"),
                    get_problem_command("https://example.com/beta"),
                    get_problem_command("https://example.com/gamma"),
                    get_problem_command("https://example.com/delta"),
                ]
            ),
        ],
        {
            "http://example.com/alpha",
            "http://example.com/beta",
            "http://example.com/gamma",
            "http://example.com/delta",
            "https://example.com/alpha",
            "https://example.com/beta",
            "https://example.com/gamma",
            "https://example.com/delta",
        },
    ),
    (
        [
            VerificationFile(
                verification=[
                    get_problem_command("http://example.com/alpha"),
                    get_problem_command("http://example.com/beta"),
                    get_problem_command("http://example.com/gamma"),
                    get_problem_command("http://example.com/delta"),
                    get_problem_command("https://example.com/alpha"),
                ]
            ),
            "https://example.com/alpha",
            "https://example.com/beta",
            "https://example.com/beta",
            "https://example.com/gamma",
            "https://example.com/delta",
        ],
        {
            "http://example.com/alpha",
            "http://example.com/beta",
            "http://example.com/gamma",
            "http://example.com/delta",
            "https://example.com/alpha",
            "https://example.com/beta",
            "https://example.com/gamma",
            "https://example.com/delta",
        },
    ),
]


@pytest.mark.parametrize(
    "input, expected",
    test_parse_urls_params,
    ids=range(len(test_parse_urls_params)),
)
def test_parse_urls(
    input: _SomeUrlOrVerificationFile,
    expected: set[str],
):
    assert set(parse_urls(input)) == expected



# Python Version: 3.x
import glob
import pathlib
from subprocess import CompletedProcess
from typing import TYPE_CHECKING, Callable, Iterator, Optional, Union

from competitive_verifier.exec import exec_command as _exec_command
from competitive_verifier.util import read_text_normalized  # noqa # pyright: ignore

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

    _StrOrListStr = Union[str, list[str]]


def glob_with_predicate(pred: Callable[[pathlib.Path], bool]) -> Iterator[pathlib.Path]:
    """glob_with_basename iterates files whose basenames satisfy the predicate.

    This function ignores hidden directories and hidden files, whose names start with dot `.` letter.
    """
    return filter(pred, map(pathlib.Path, glob.glob("**", recursive=True)))


def exec_command(
    command: "_StrOrListStr",
    *,
    check: bool = False,
    env: Optional[dict[str, str]] = None,
    cwd: Optional["StrOrBytesPath"] = None,
) -> CompletedProcess[bytes]:
    return _exec_command(
        command,
        check=check,
        env=env,
        cwd=cwd,
        text=False,
        capture_output=True,
        group_log=False,
    )



# pyright: reportUnknownVariableType=false, reportUnknownMemberType=false
# Python Version: 3.x
import concurrent.futures
import functools
import os
import pathlib
import platform
import sys
from logging import getLogger
from typing import Sequence

import importlab.environment
import importlab.fs
import importlab.graph

from competitive_verifier.oj.verify.models import Language, LanguageEnvironment

logger = getLogger(__name__)


class PythonLanguageEnvironment(LanguageEnvironment):
    @property
    def name(self) -> str:
        return "Python"

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        python_path = os.getenv("PYTHONPATH")
        if python_path:
            python_path = basedir.resolve().as_posix() + os.pathsep + python_path
        else:
            python_path = basedir.resolve().as_posix()
        return f"env PYTHONPATH={python_path} python {path}"


@functools.lru_cache(maxsize=None)
def _python_list_depending_files(
    path: pathlib.Path, basedir: pathlib.Path
) -> list[pathlib.Path]:
    # compute the dependency graph of the `path`
    env = importlab.environment.Environment(
        importlab.fs.Path([importlab.fs.OSFileSystem(str(basedir.resolve()))]),
        (sys.version_info.major, sys.version_info.minor),
    )
    try:
        executor = concurrent.futures.ThreadPoolExecutor()
        future = executor.submit(
            importlab.graph.ImportGraph.create,  # pyright: ignore[reportUnknownArgumentType]
            env,
            [str(path)],
            True,
        )
        if platform.uname().system == "Windows":
            timeout = 5.0  # 1.0 sec causes timeout on CI using Windows
        else:
            timeout = 1.0
        res_graph = future.result(timeout=timeout)
    except concurrent.futures.TimeoutError as e:
        raise RuntimeError(
            f"Failed to analyze the dependency graph (timeout): {path}"
        ) from e
    try:
        node_deps_pairs: list[tuple[str, list[str]]] = res_graph.deps_list()
    except Exception as e:
        raise RuntimeError(
            f"Failed to analyze the dependency graph (circular imports?): {path}"
        ) from e
    logger.debug("the dependency graph of %s: %s", str(path), node_deps_pairs)

    # collect Python files which are depended by the `path` and under `basedir`
    res_deps: list[pathlib.Path] = []
    res_deps.append(path.resolve())
    for node_, deps_ in node_deps_pairs:
        node = pathlib.Path(node_)
        deps = list(map(pathlib.Path, deps_))
        if node.resolve() == path.resolve():
            for dep in deps:
                if basedir.resolve() in dep.resolve().parents:
                    res_deps.append(dep.resolve())
            break
    return list(set(res_deps))


class PythonLanguage(Language):
    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        return _python_list_depending_files(path.resolve(), basedir)

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> Sequence[PythonLanguageEnvironment]:
        return [PythonLanguageEnvironment()]



from enum import Enum


class ResultStatus(str, Enum):
    SUCCESS = "success"
    FAILURE = "failure"
    SKIPPED = "skipped"

    @property
    def status(self) -> "ResultStatus":
        return self



import contextlib
import os
import pathlib
from typing import Any, TypedDict

import onlinejudge.service.atcoder as atcoder
import onlinejudge.service.library_checker as library_checker
import onlinejudge.service.yukicoder as yukicoder
import onlinejudge.type
import pytest
import requests
from pytest_mock import MockerFixture
from pytest_mock.plugin import MockType

from competitive_verifier.download.main import run_impl as download


class MockProblem(TypedDict):
    download_system_cases: MockType
    download_sample_cases: MockType


@pytest.fixture
def mock_problem(mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.chdir(pathlib.Path(__file__).parent)
    mocker.patch.dict(
        os.environ, {"YUKICODER_TOKEN": "YKTK", "DROPBOX_TOKEN": "DBTK"}, clear=True
    )
    mocker.patch(
        "competitive_verifier.oj.tools.test_command.get_cache_directory",
        return_value=pathlib.Path("/bar/baz/online-judge-tools"),
    )
    mocker.patch(
        "competitive_verifier.oj.tools.download_command.get_checker_path",
        return_value=None,
    )

    @contextlib.contextmanager
    def new_session_with_our_user_agent(*, path: pathlib.Path):
        sess = requests.Session()
        sess.headers = {}
        yield sess

    mocker.patch(
        "onlinejudge_command.utils.new_session_with_our_user_agent",
        side_effect=new_session_with_our_user_agent,
    )

    return {
        problem_type: MockProblem(
            download_system_cases=mocker.patch.object(
                problem_type,
                "download_system_cases",
                return_value=[],
            ),
            download_sample_cases=mocker.patch.object(
                problem_type,
                "download_sample_cases",
                return_value=[],
            ),
        )
        for problem_type in [
            library_checker.LibraryCheckerProblem,
            yukicoder.YukicoderProblem,
            atcoder.AtCoderProblem,
        ]
    }


@pytest.fixture
def mkdir(mocker: MockerFixture):
    return mocker.patch.object(pathlib.Path, "mkdir", autospec=True)


def test_oj_download(
    mkdir: MockType,
    mock_problem: dict[type[onlinejudge.type.Problem], MockProblem],
):
    download(
        input=[
            "https://judge.yosupo.jp/problem/aplusb",
            "https://judge.yosupo.jp/problem/aplusb",
            "https://yukicoder.me/problems/no/3040",
            "https://atcoder.jp/contests/abc322/tasks/abc322_a",
        ]
    )

    assert mkdir.call_count == 3
    assert sorted(call[0][0] for call in mkdir.call_args_list) == [
        pathlib.Path(
            ".competitive-verifier/cache/problems/20c21841818196666af22f1bfb3dbd3e"
        ),
        pathlib.Path(
            ".competitive-verifier/cache/problems/84d0ff7b028fc0f8cf7973b22fee1634"
        ),
        pathlib.Path(
            ".competitive-verifier/cache/problems/8e3916c7805235eb07ec2a58660d89c6"
        ),
    ]

    for m in mock_problem.values():
        m["download_sample_cases"].assert_not_called()
        m["download_system_cases"].assert_called_once()

    class SessionData(TypedDict):
        headers: dict[str, Any]
        cookies: dict[str, Any]

    calls: dict[type[onlinejudge.type.Problem], SessionData] = {
        k: {
            "headers": v["download_system_cases"].call_args[1]["session"].headers,
            "cookies": v["download_system_cases"]
            .call_args[1]["session"]
            .cookies.get_dict(),
        }
        for k, v in mock_problem.items()
    }

    assert calls == {
        library_checker.LibraryCheckerProblem: {"headers": {}, "cookies": {}},
        yukicoder.YukicoderProblem: {
            "headers": {"Authorization": "Bearer YKTK"},
            "cookies": {},
        },
        atcoder.AtCoderProblem: {
            "headers": {"Authorization": "Bearer DBTK"},
            "cookies": {},
        },
    }



import argparse
from typing import Optional

import pytest

from competitive_verifier.error import VerifierError
from competitive_verifier.verify.main import argument as argument_verify
from competitive_verifier.verify.main import get_split_state
from competitive_verifier.verify.verifier import SplitState

test_get_split_state_params = [
    (None, None, None),
    (5, 0, SplitState(size=5, index=0)),
    (5, 1, SplitState(size=5, index=1)),
    (5, 2, SplitState(size=5, index=2)),
    (5, 3, SplitState(size=5, index=3)),
    (5, 4, SplitState(size=5, index=4)),
]


@pytest.mark.parametrize(
    "size, index, expected",
    test_get_split_state_params,
    ids=range(len(test_get_split_state_params)),
)
def test_get_split_state(
    size: Optional[int],
    index: Optional[int],
    expected: Optional[SplitState],
):
    assert get_split_state(size, index) == expected


get_split_state_error_params = {
    "No split index": (
        ["--verify-json", "verify.json", "--split", "2"],
        "--split argument requires --split-index argument.",
    ),
    "No split": (
        ["--verify-json", "verify.json", "--split-index", "2"],
        "--split-index argument requires --split argument.",
    ),
    "split index": (
        ["--verify-json", "verify.json", "--split-index", "5", "--split", "5"],
        "--split-index must be greater than 0 and less than --split.",
    ),
    "split index negative": (
        ["--verify-json", "verify.json", "--split-index", "-1", "--split", "5"],
        "--split-index must be greater than 0 and less than --split.",
    ),
    "split zero": (
        ["--verify-json", "verify.json", "--split-index", "1", "--split", "0"],
        "--split must be greater than 0.",
    ),
}


@pytest.mark.parametrize(
    "args, message",
    get_split_state_error_params.values(),
    ids=get_split_state_error_params.keys(),
)
def test_get_split_state_error(args: list[str], message: str):
    parser = argument_verify(argparse.ArgumentParser())
    parsed = parser.parse_args(args)

    with pytest.raises(VerifierError) as e:
        get_split_state(parsed.split, parsed.split_index)

    assert e.value.message == message



import shutil
from typing import Any

import pytest

from ..types import ConfigDirSetter, FilePaths
from .integration_data import IntegrationData


class JavaData(IntegrationData):
    def __init__(
        self,
        monkeypatch: pytest.MonkeyPatch,
        set_config_dir: ConfigDirSetter,
        file_paths: FilePaths,
    ) -> None:
        super().__init__(monkeypatch, set_config_dir, file_paths)

    def check_envinronment(self) -> bool:
        return bool(shutil.which("javac"))

    def expected_verify_json(self) -> dict[str, Any]:
        return dict(
            {
                "files": {
                    "examples/Aplutb_test.java": {
                        "additonal_sources": [],
                        "dependencies": [
                            "examples/Aplutb_test.java",
                            "examples/HelloWorld.java",
                            "examples/HelloWorld_test.java",
                        ],
                        "document_attributes": {
                            "PROBLEM": "https://judge.yosupo.jp/problem/aplusb"
                        },
                        "verification": [
                            {
                                "command": "java examples.Aplutb_test",
                                "compile": f"javac {self.targets_path/'examples/Aplutb_test.java'}",
                                "name": "Java",
                                "problem": "https://judge.yosupo.jp/problem/aplusb",
                                "type": "problem",
                            }
                        ],
                    },
                    "examples/HelloWorld.java": {
                        "additonal_sources": [],
                        "dependencies": [
                            "examples/Aplutb_test.java",
                            "examples/HelloWorld.java",
                            "examples/HelloWorld_test.java",
                        ],
                        "document_attributes": {},
                        "verification": [],
                    },
                    "examples/HelloWorld_test.java": {
                        "additonal_sources": [],
                        "dependencies": [
                            "examples/Aplutb_test.java",
                            "examples/HelloWorld.java",
                            "examples/HelloWorld_test.java",
                        ],
                        "document_attributes": {
                            "PROBLEM": "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A"
                        },
                        "verification": [
                            {
                                "command": "java examples.HelloWorld_test",
                                "compile": f"javac {self.targets_path/'examples/HelloWorld_test.java'}",
                                "name": "Java",
                                "problem": "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A",
                                "type": "problem",
                            }
                        ],
                    },
                },
            },
        )

    def expected_verify_result(self) -> dict[str, Any]:
        return {
            "files": {
                "examples/Aplutb_test.java": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 5673.0,
                            "heaviest": 12.0,
                            "last_execution_time": "2029-07-26T03:21:57.010000-11:00",
                            "slowest": 567.0,
                            "status": "success",
                            "testcases": [
                                {
                                    "elapsed": 6.24,
                                    "memory": 23.23,
                                    "name": "example_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 8.85,
                                    "memory": 85.03,
                                    "name": "example_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 3.51,
                                    "memory": 66.88,
                                    "name": "random_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 3.94,
                                    "memory": 18.9,
                                    "name": "random_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 0.27,
                                    "memory": 9.41,
                                    "name": "random_02",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.34,
                                    "memory": 26.78,
                                    "name": "random_03",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 0.34,
                                    "memory": 2.03,
                                    "name": "random_04",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 3.24,
                                    "memory": 36.55,
                                    "name": "random_05",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.52,
                                    "memory": 51.34,
                                    "name": "random_06",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 2.22,
                                    "memory": 30.05,
                                    "name": "random_07",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 0.38,
                                    "memory": 35.4,
                                    "name": "random_08",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 6.03,
                                    "memory": 68.16,
                                    "name": "random_09",
                                    "status": "AC",
                                },
                            ],
                            "verification_name": "Java",
                        }
                    ],
                },
                "examples/HelloWorld_test.java": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 8002.0,
                            "heaviest": 302.0,
                            "last_execution_time": "2007-06-15T09:14:19.970000+10:00",
                            "slowest": 800.0,
                            "status": "success",
                            "testcases": [
                                {
                                    "elapsed": 3.73,
                                    "memory": 70.0,
                                    "name": "judge_data",
                                    "status": "AC",
                                }
                            ],
                            "verification_name": "Java",
                        }
                    ],
                },
            },
            "total_seconds": 2547.12,
        }



import datetime
import pathlib
from hashlib import md5
from typing import Iterable


def md5_number(seed: bytes):
    return int(md5(seed).hexdigest(), 16)


def dummy_commit_time(files: Iterable[pathlib.Path]) -> datetime.datetime:
    md5 = md5_number(b"".join(sorted(f.as_posix().encode() for f in files)))

    return datetime.datetime.fromtimestamp(
        md5 % 300000000000 / 100,
        tz=datetime.timezone(datetime.timedelta(hours=md5 % 25 - 12)),
    )



import contextlib
import inspect
import os
import pathlib
import shutil
from typing import Optional

import pytest
import requests
from pytest_mock import MockerFixture

from .data.cpp import CppWithConfigData, CppWithoutConfigData
from .data.go import GoWithConfigData, GoWithoutConfigData
from .data.integration_data import IntegrationData
from .data.java import JavaData
from .data.rust import RustWithoutConfigData
from .data.user_defined_and_python import UserDefinedAndPythonData
from .types import ConfigDirSetter, FilePaths
from .utils import dummy_commit_time


@pytest.fixture(scope="session")
def check_necessary_commands() -> Optional[str]:
    git_path = shutil.which("git")
    if not git_path:
        raise Exception("The integration test needs command")
    if shutil.which("env"):
        return None
    if os.name != "nt":
        raise Exception("The integration test needs env command")

    for git_dir in pathlib.Path(git_path).parents:
        search = git_dir / "usr/bin"
        if search.is_dir():
            return str(search)
    raise Exception("The integration test needs env command")


@pytest.fixture()
def additional_path(
    monkeypatch: pytest.MonkeyPatch,
    check_necessary_commands: Optional[str],
):
    if check_necessary_commands:
        monkeypatch.setenv("PATH", check_necessary_commands, prepend=os.pathsep)


@pytest.fixture(scope="session")
def file_paths(request: pytest.FixtureRequest) -> FilePaths:
    root = pathlib.Path(__file__).parent.parent.parent / "integration_test_data"
    dest_root = root / "dst_dir"
    assert root.exists()

    if dest_root.is_dir() and not request.config.getoption("--use-prev-dest"):
        shutil.rmtree(dest_root)

    return FilePaths(
        root=root,
        dest_root=dest_root,
    )


@pytest.fixture
def set_config_dir(mocker: MockerFixture, file_paths: FilePaths) -> ConfigDirSetter:
    def _set_config_dir(name: str):
        d = file_paths.dest_root / "config" / (name or inspect.stack()[1].function)

        mocker.patch.dict(
            os.environ,
            {"COMPETITIVE_VERIFY_CONFIG_PATH": d.as_posix()},
        )
        return d

    return _set_config_dir


@pytest.fixture(autouse=True)
def setenv(
    mocker: MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    file_paths: FilePaths,
):
    monkeypatch.chdir(file_paths.root)

    mocker.patch(
        "competitive_verifier.git.get_commit_time",
        side_effect=dummy_commit_time,
    )

    @contextlib.contextmanager
    def new_session_with_our_user_agent(*, path: pathlib.Path):
        sess = requests.Session()
        sess.headers = {}
        yield sess

    mocker.patch(
        "onlinejudge_command.utils.new_session_with_our_user_agent",
        side_effect=new_session_with_our_user_agent,
    )


@pytest.fixture
def user_defined_and_python_data(
    monkeypatch: pytest.MonkeyPatch,
    file_paths: FilePaths,
    set_config_dir: ConfigDirSetter,
) -> UserDefinedAndPythonData:
    return UserDefinedAndPythonData(monkeypatch, set_config_dir, file_paths)


@pytest.fixture(
    params=[
        UserDefinedAndPythonData,
        GoWithConfigData,
        GoWithoutConfigData,
        JavaData,
        CppWithConfigData,
        CppWithoutConfigData,
        RustWithoutConfigData,
    ]
)
def integration_data(
    request: pytest.FixtureRequest,
    monkeypatch: pytest.MonkeyPatch,
    file_paths: FilePaths,
    set_config_dir: ConfigDirSetter,
) -> IntegrationData:
    cls = request.param
    assert issubclass(cls, IntegrationData)
    d = cls(monkeypatch, set_config_dir, file_paths)
    if not d.check_envinronment():
        pytest.skip(f"No envinronment {cls.__name__}")
    return d



import shutil
from typing import Any, Optional

import pytest

from ..types import ConfigDirSetter, FilePaths
from .integration_data import IntegrationData


class CppWithoutConfigData(IntegrationData):
    def __init__(
        self,
        monkeypatch: pytest.MonkeyPatch,
        set_config_dir: ConfigDirSetter,
        file_paths: FilePaths,
    ) -> None:
        super().__init__(monkeypatch, set_config_dir, file_paths)

    def check_envinronment(self) -> bool:
        return bool(shutil.which("g++") and shutil.which("clang++"))

    @classmethod
    @property
    def input_name(cls) -> str:
        return "CppData"

    def expected_verify_json(self) -> dict[str, Any]:
        return dict(
            {
                "files": {
                    "aplusb.hpp": {
                        "additonal_sources": [
                            {
                                "name": "bundled",
                                "path": str(
                                    self.config_dir_path / "bundled/aplusb.hpp"
                                ),
                            }
                        ],
                        "dependencies": ["aplusb.hpp"],
                        "document_attributes": {
                            "*NOT_SPECIAL_COMMENTS*": "",
                            "links": [],
                        },
                        "verification": [],
                    },
                    "aplusb.test.cpp": {
                        "additonal_sources": [
                            {
                                "name": "bundled",
                                "path": str(
                                    self.config_dir_path / "bundled/aplusb.test.cpp"
                                ),
                            }
                        ],
                        "dependencies": ["aplusb.hpp", "aplusb.test.cpp", "macros.hpp"],
                        "document_attributes": {
                            "*NOT_SPECIAL_COMMENTS*": "",
                            "PROBLEM": "https://judge.yosupo.jp/problem/aplusb",
                            "links": ["https://judge.yosupo.jp/problem/aplusb"],
                        },
                        "verification": [
                            {
                                "command": f"{self.config_dir_path/'cache/problems/8e3916c7805235eb07ec2a58660d89c6/a.out'}",
                                "compile": "/usr/bin/g++ "
                                "--std=c++17 -O2 -Wall -g -I "
                                f"{self.targets_path} "
                                "-o "
                                f"{self.config_dir_path/'cache/problems/8e3916c7805235eb07ec2a58660d89c6/a.out'} "
                                "aplusb.test.cpp",
                                "name": "g++",
                                "problem": "https://judge.yosupo.jp/problem/aplusb",
                                "type": "problem",
                            },
                            {
                                "command": f"{self.config_dir_path/'cache/problems/8e3916c7805235eb07ec2a58660d89c6/a.out'}",
                                "compile": "/usr/bin/clang++ "
                                "--std=c++17 -O2 -Wall -g -I "
                                f"{self.targets_path} "
                                "-o "
                                f"{self.config_dir_path/'cache/problems/8e3916c7805235eb07ec2a58660d89c6/a.out'} "
                                "aplusb.test.cpp",
                                "name": "clang++",
                                "problem": "https://judge.yosupo.jp/problem/aplusb",
                                "type": "problem",
                            },
                        ],
                    },
                    "macros.hpp": {
                        "additonal_sources": [
                            {
                                "name": "bundled",
                                "path": str(
                                    self.config_dir_path / "bundled/macros.hpp"
                                ),
                            }
                        ],
                        "dependencies": ["macros.hpp"],
                        "document_attributes": {
                            "*NOT_SPECIAL_COMMENTS*": "",
                            "links": [],
                        },
                        "verification": [],
                    },
                },
            },
        )

    def expected_verify_result(self) -> dict[str, Any]:
        return {
            "files": {
                "aplusb.test.cpp": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 7135.0,
                            "heaviest": 8.0,
                            "last_execution_time": "2039-07-14T04:50:17.220000+10:00",
                            "slowest": 713.0,
                            "status": "success",
                            "testcases": [
                                {
                                    "elapsed": 8.65,
                                    "memory": 44.22,
                                    "name": "example_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 1.7,
                                    "memory": 33.42,
                                    "name": "example_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 4.22,
                                    "memory": 72.01,
                                    "name": "random_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 1.92,
                                    "memory": 37.04,
                                    "name": "random_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.54,
                                    "memory": 35.13,
                                    "name": "random_02",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 7.19,
                                    "memory": 75.36,
                                    "name": "random_03",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 7.27,
                                    "memory": 99.94,
                                    "name": "random_04",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 7.22,
                                    "memory": 11.11,
                                    "name": "random_05",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 6.69,
                                    "memory": 31.8,
                                    "name": "random_06",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 2.96,
                                    "memory": 2.22,
                                    "name": "random_07",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 3.35,
                                    "memory": 7.1,
                                    "name": "random_08",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 5.37,
                                    "memory": 80.29,
                                    "name": "random_09",
                                    "status": "AC",
                                },
                            ],
                            "verification_name": "g++",
                        },
                        {
                            "elapsed": 8550.0,
                            "heaviest": 236.0,
                            "last_execution_time": "1974-07-10T15:38:07.610000-01:00",
                            "slowest": 855.0,
                            "status": "success",
                            "testcases": [
                                {
                                    "elapsed": 8.88,
                                    "memory": 32.57,
                                    "name": "example_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 2.86,
                                    "memory": 36.07,
                                    "name": "example_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.11,
                                    "memory": 66.14,
                                    "name": "random_00",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 0.19,
                                    "memory": 47.34,
                                    "name": "random_01",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 5.42,
                                    "memory": 56.25,
                                    "name": "random_02",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 4.08,
                                    "memory": 21.26,
                                    "name": "random_03",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.75,
                                    "memory": 66.28,
                                    "name": "random_04",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.05,
                                    "memory": 29.79,
                                    "name": "random_05",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 0.51,
                                    "memory": 97.67,
                                    "name": "random_06",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 0.15,
                                    "memory": 20.61,
                                    "name": "random_07",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.07,
                                    "memory": 96.03,
                                    "name": "random_08",
                                    "status": "AC",
                                },
                                {
                                    "elapsed": 9.86,
                                    "memory": 41.06,
                                    "name": "random_09",
                                    "status": "AC",
                                },
                            ],
                            "verification_name": "clang++",
                        },
                    ],
                }
            },
            "total_seconds": 1312.56,
        }


class CppWithConfigData(CppWithoutConfigData):
    def __init__(
        self,
        monkeypatch: pytest.MonkeyPatch,
        set_config_dir: ConfigDirSetter,
        file_paths: FilePaths,
    ) -> None:
        super().__init__(monkeypatch, set_config_dir, file_paths)

    @property
    def config_path(self) -> Optional[str]:
        return (self.targets_path / "config.toml").as_posix()

    def expected_verify_json(self) -> dict[str, Any]:
        return dict(
            {
                "files": {
                    "aplusb.hpp": {
                        "additonal_sources": [
                            {
                                "name": "bundled",
                                "path": str(
                                    self.config_dir_path / "bundled/aplusb.hpp"
                                ),
                            }
                        ],
                        "dependencies": ["aplusb.hpp"],
                        "document_attributes": {
                            "*NOT_SPECIAL_COMMENTS*": "",
                            "links": [],
                        },
                        "verification": [],
                    },
                    "aplusb.test.cpp": {
                        "additonal_sources": [
                            {
                                "name": "bundled",
                                "path": str(
                                    self.config_dir_path / "bundled/aplusb.test.cpp"
                                ),
                            }
                        ],
                        "dependencies": ["aplusb.hpp", "aplusb.test.cpp", "macros.hpp"],
                        "document_attributes": {
                            "*NOT_SPECIAL_COMMENTS*": "",
                            "PROBLEM": "https://judge.yosupo.jp/problem/aplusb",
                            "links": ["https://judge.yosupo.jp/problem/aplusb"],
                        },
                        "verification": [
                            {
                                "command": f"{self.config_dir_path/'cache/problems/8e3916c7805235eb07ec2a58660d89c6/a.out'}",
                                "compile": "g++ "
                                "--std=c++17 -Wall -g -I "
                                f"{self.targets_path} "
                                "-o "
                                f"{self.config_dir_path/'cache/problems/8e3916c7805235eb07ec2a58660d89c6/a.out'} "
                                "aplusb.test.cpp",
                                "name": "g++",
                                "problem": "https://judge.yosupo.jp/problem/aplusb",
                                "type": "problem",
                            },
                            {
                                "command": f"{self.config_dir_path/'cache/problems/8e3916c7805235eb07ec2a58660d89c6/a.out'}",
                                "compile": "clang++ "
                                "--std=c++17 -Wall -g -I "
                                f"{self.targets_path} "
                                "-o "
                                f"{self.config_dir_path/'cache/problems/8e3916c7805235eb07ec2a58660d89c6/a.out'} "
                                "aplusb.test.cpp",
                                "name": "clang++",
                                "problem": "https://judge.yosupo.jp/problem/aplusb",
                                "type": "problem",
                            },
                        ],
                    },
                    "macros.hpp": {
                        "additonal_sources": [
                            {
                                "name": "bundled",
                                "path": str(
                                    self.config_dir_path / "bundled/macros.hpp"
                                ),
                            }
                        ],
                        "dependencies": ["macros.hpp"],
                        "document_attributes": {
                            "*NOT_SPECIAL_COMMENTS*": "",
                            "links": [],
                        },
                        "verification": [],
                    },
                },
            },
        )
